# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2025 ODC Contributors
# SPDX-License-Identifier: Apache-2.0

# We often have one-arg-per column, so these checks aren't so useful.
# pylint: disable=too-many-arguments,too-many-public-methods,too-many-lines

# SQLAlchemy queries require "column == None", not "column is None" due to operator overloading:
# pylint: disable=singleton-comparison

"""
Persistence API implementation for postgis.
"""

import datetime
import json
import logging
import uuid
from collections.abc import Iterable, Iterator, Sequence
from typing import Any
from typing import cast as type_cast

from odc.geo import CRS, Geometry
from sqlalchemy import (
    and_,
    cast,
    column,
    delete,
    func,
    or_,
    select,
    text,
    update,
)
from sqlalchemy.dialects.postgresql import INTERVAL, insert
from sqlalchemy.exc import IntegrityError
from sqlalchemy.sql.expression import Select
from typing_extensions import override

from datacube.index.abstract import DSID
from datacube.index.fields import OrExpression
from datacube.model import Range
from datacube.model.fields import Expression
from datacube.model.lineage import LineageDirection, LineageRelation
from datacube.utils.uris import split_uri

from ...utils.changes import Offset
from . import _core
from ._fields import (
    DateDocField,
    DateRangeDocField,
    NativeField,
    PgExpression,
    PgField,
    SimpleDocField,
    UnindexableValue,
    parse_fields,
)
from ._schema import (
    Dataset,
    DatasetHome,
    DatasetLineage,
    MetadataType,
    Product,
    search_field_index_map,
    search_field_indexes,
)
from ._spatial import (
    extract_geometry_from_eo3_projection,
    generate_dataset_spatial_values,
    geom_alchemy,
)
from .sql import escape_pg_identifier

_LOG: logging.Logger = logging.getLogger(__name__)


# Make a function because it's broken
def _dataset_select_fields() -> tuple:
    return tuple(f.alchemy_expression for f in _dataset_fields())


def _base_known_fields() -> dict:
    fields = get_native_fields().copy()
    fields["archived"] = NativeField(
        "archived", "Archived date", Dataset.__table__.c.archived
    )
    return fields


def _dataset_fields() -> tuple:
    native_flds = get_native_fields()
    return (
        native_flds["id"],
        native_flds["indexed_time"],
        native_flds["indexed_by"],
        native_flds["product_id"],
        native_flds["metadata_type_id"],
        native_flds["metadata_doc"],
        NativeField("archived", "Archived date", Dataset.__table__.c.archived),
        native_flds["uri"],
    )


def _dataset_bulk_select_fields() -> tuple:
    return (Dataset.product_ref, Dataset.metadata_doc, Dataset.uri)


def get_native_fields() -> dict[str, PgField]:
    # Native fields (hard-coded into the schema)
    fields: dict[str, PgField] = {
        "id": NativeField("id", "Dataset UUID", Dataset.__table__.c.id),
        "indexed_time": NativeField(
            "indexed_time", "When dataset was indexed", Dataset.__table__.c.added
        ),
        "indexed_by": NativeField(
            "indexed_by", "User who indexed the dataset", Dataset.__table__.c.added_by
        ),
        "product": NativeField(
            "product",
            "Product name",
            Product.__table__.c.name,
            join_clause=(Product.id == Dataset.product_ref),
        ),
        "product_id": NativeField(
            "product_id", "ID of a dataset type", Dataset.__table__.c.product_ref
        ),
        "metadata_type": NativeField(
            "metadata_type",
            "Metadata type name of dataset",
            MetadataType.__table__.c.name,
            join_clause=(MetadataType.id == Dataset.metadata_type_ref),
        ),
        "metadata_type_id": NativeField(
            "metadata_type_id",
            "ID of a metadata type",
            Dataset.__table__.c.metadata_type_ref,
        ),
        "metadata_doc": NativeField(
            "metadata_doc",
            "Full metadata document",
            Dataset.metadata_doc,  # type: ignore[arg-type]
        ),
        "uri": NativeField(
            "uri",
            "Dataset URI",
            Dataset.__table__.c.uri_body,
            alchemy_expression=Dataset.uri,
        ),
    }
    return fields


def mk_simple_offset_field(
    field_name: str, description: str, offset: Offset
) -> SimpleDocField:
    return SimpleDocField(
        name=field_name,
        description=description,
        alchemy_column=Dataset.metadata_doc,
        indexed=False,
        offset=offset,
    )


def get_dataset_fields(metadata_type_definition: dict[str, Any]) -> dict[str, Any]:
    dataset_section = metadata_type_definition["dataset"]

    fields = get_native_fields()
    # "Fixed fields" (not dynamic: defined in metadata type schema)
    fields.update(
        {
            "creation_time": DateDocField(
                "creation_time",
                "Time when dataset was created (processed)",
                Dataset.metadata_doc,
                False,
                offset=dataset_section.get("creation_dt") or ("creation_dt",),
            ),
            "format": mk_simple_offset_field(
                "format",
                "File format (GeoTiff, NetCDF)",
                dataset_section.get("format") or ("format", "name"),
            ),
            "label": mk_simple_offset_field(
                "label", "Label", dataset_section.get("label") or ("label",)
            ),
        }
    )

    # noinspection PyTypeChecker
    fields.update(parse_fields(dataset_section["search_fields"], Dataset.metadata_doc))
    return fields


def non_native_fields(mdt_metadata) -> dict:
    return {
        name: field
        for name, field in get_dataset_fields(mdt_metadata).items()
        if not isinstance(field, NativeField)
    }


def extract_dataset_search_fields(ds_metadata, mdt_metadata) -> dict:
    """
    :param ds_metadata: A Dataset metadata document
    :param mdt_metadata: The corresponding metadata-type definition document

    :return: A dictionary mapping search field names to (type_name, value) tuples.
    """
    return extract_dataset_fields(ds_metadata, non_native_fields(mdt_metadata))


def extract_dataset_fields(ds_metadata, fields) -> dict:
    """
    :param ds_metadata: A Dataset metadata document
    :param fields: A dictionary of field names to Field objects

    :return: A dictionary mapping search field names to (type_name, value) tuples.
    """
    result = {}
    for field_name, field in fields.items():
        try:
            fld_type = field.type_name
            raw_val = field.extract(ds_metadata)
            sqla_val = field.search_value_to_alchemy(raw_val)
            result[field_name] = (fld_type, sqla_val)
        except UnindexableValue:
            continue
    return result


# Min/Max aggregating time fields for temporal_extent methods
time_min = DateDocField(
    "acquisition_time_min",
    "Min of time when dataset was acquired",
    Dataset.metadata_doc,
    False,  # is it indexed
    offset=[("properties", "dtr:start_datetime"), ("properties", "datetime")],
    selection="least",
)


time_max = DateDocField(
    "acquisition_time_max",
    "Max of time when dataset was acquired",
    Dataset.metadata_doc,
    False,  # is it indexed
    offset=[("properties", "dtr:end_datetime"), ("properties", "datetime")],
    selection="greatest",
)


class PostgisDbAPI:
    def __init__(self, parentdb, connection) -> None:
        self._db = parentdb
        self._connection = connection
        self._sqla_txn = None

    @property
    def in_transaction(self):
        return self._connection.in_transaction()

    def begin(self) -> None:
        self._connection.execution_options(isolation_level="REPEATABLE READ")
        self._sqla_txn = self._connection.begin()

    def _end_transaction(self) -> None:
        self._sqla_txn = None
        self._connection.execution_options(isolation_level="AUTOCOMMIT")

    def commit(self) -> None:
        self._sqla_txn.commit()  # type: ignore[attr-defined]
        self._end_transaction()

    def rollback(self) -> None:
        self._sqla_txn.rollback()  # type: ignore[attr-defined]
        self._end_transaction()

    def execute(self, command):
        return self._connection.execute(command)

    def insert_dataset(self, metadata_doc, dataset_id, product_id) -> bool:
        """
        Insert dataset if not already indexed.
        :return: whether it was inserted
        """
        metadata_subquery = (
            select(Product.metadata_type_ref)
            .where(Product.id == product_id)
            .scalar_subquery()
        )
        ret = self._connection.execute(
            insert(Dataset.__table__)  # type: ignore[arg-type]
            .values(
                id=dataset_id,
                product_ref=product_id,
                metadata=metadata_doc,
                metadata_type_ref=metadata_subquery,
            )
            .on_conflict_do_nothing(index_elements=["id"])
        )
        return ret.rowcount > 0

    def insert_dataset_bulk(self, values) -> tuple:
        requested = len(values)
        res = self._connection.execute(
            insert(Dataset.__table__),  # type: ignore[arg-type]
            values,
        )
        return res.rowcount, requested - res.rowcount

    def update_dataset(self, metadata_doc, dataset_id, product_id) -> bool:
        """
        Update dataset
        """
        res = self._connection.execute(
            update(Dataset.__table__)  # type: ignore[arg-type]
            .returning(Dataset.__table__.c.id)
            .where(Dataset.__table__.c.id == dataset_id)
            .where(Dataset.__table__.c.product_ref == product_id)
            .values(metadata=metadata_doc)
        )
        return res.rowcount > 0

    def insert_dataset_location(self, dataset_id: DSID, uri: str) -> bool:
        """
        Add a location to a dataset if it is not already recorded.

        Returns True if success, False if this location already existed
        """
        scheme, body = split_uri(uri)

        r = self._connection.execute(
            update(Dataset.__table__)  # type: ignore[arg-type]
            .returning(Dataset.uri)
            .where(Dataset.__table__.c.id == dataset_id)
            .values(uri_scheme=scheme, uri_body=body)
        )

        return r.rowcount > 0

    def insert_dataset_search(self, search_table, dataset_id: DSID, key, value) -> bool:
        """
        Add/update a search field index entry for a dataset

        Returns True on success
        """
        if isinstance(value, Range):
            value = list(value)
        r = self._connection.execute(
            insert(search_table)
            .values(
                dataset_ref=dataset_id,
                search_key=key,
                search_val=value,
            )
            .on_conflict_do_update(
                index_elements=[search_table.dataset_ref, search_table.search_key],
                set_={"search_val": value},
            )
        )
        return r.rowcount > 0

    def insert_dataset_search_bulk(self, search_type: str, values):
        search_table = search_field_index_map[search_type]
        r = self._connection.execute(insert(search_table).values(values))
        return r.rowcount

    def insert_dataset_spatial(
        self, dataset_id: DSID, crs: CRS, extent: Geometry
    ) -> bool:
        """
        Add/update a spatial index entry for a dataset

        Returns True on success
        """
        values = generate_dataset_spatial_values(dataset_id, crs, extent)
        if values is None:
            return False
        SpatialIndex = self._db.spatial_index(crs)  # noqa: N806
        r = self._connection.execute(
            insert(SpatialIndex)
            .values(**values)
            .on_conflict_do_update(
                index_elements=[SpatialIndex.dataset_ref],
                set_={"extent": values["extent"]},
            )
        )
        return r.rowcount > 0

    def insert_dataset_spatial_bulk(self, crs, values):
        SpatialIndex = self._db.spatial_index(crs)  # noqa: N806
        r = self._connection.execute(insert(SpatialIndex).values(values))
        return r.rowcount

    def spatial_extent(self, ids, crs):
        SpatialIndex = self._db.spatial_index(crs)  # noqa: N806
        if SpatialIndex is None:
            # Requested a CRS that has no spatial index, so use 4326 (which always has a spatial index)
            # and reproject to requested CRS.
            return self.spatial_extent(ids, CRS("epsg:4326")).to_crs(crs)
        query = (
            select(func.ST_AsGeoJSON(func.ST_Union(SpatialIndex.extent)))
            .select_from(SpatialIndex)
            .where(SpatialIndex.dataset_ref.in_(ids))
        )
        result = self._connection.execute(query)
        for r in result:
            extent_json = r[0]
            if extent_json is None:
                return None
            return Geometry(json.loads(extent_json), crs=crs)
        return None

    def contains_dataset(self, dataset_id) -> bool:
        return bool(
            self._connection.execute(
                select(Dataset.id).where(Dataset.id == dataset_id)
            ).fetchone()
        )

    def datasets_intersection(self, dataset_ids) -> list:
        """Compute set intersection: db_dataset_ids & dataset_ids"""
        return [
            ds.id
            for ds in self._connection.execute(
                select(Dataset.id).where(Dataset.id.in_(dataset_ids))
            ).fetchall()
        ]

    def get_datasets_for_location(self, uri, mode: str | None = None):
        scheme, body = split_uri(uri)

        if mode is None:
            mode = "exact" if body.count("#") > 0 else "prefix"

        if mode == "exact":
            body_query = Dataset.uri_body == body
        elif mode == "prefix":
            body_query = Dataset.uri_body.startswith(body)
        else:
            raise ValueError(f"Unsupported query mode {mode}")

        return self._connection.execute(
            select(*_dataset_select_fields()).where(
                and_(Dataset.uri_scheme == scheme, body_query)
            )
        ).fetchall()

    def all_dataset_ids(self, archived: bool | None = False):
        query = select(Dataset.id)
        if archived:
            query = query.where(Dataset.archived.is_not(None))
        elif archived is not None:
            query = query.where(Dataset.archived.is_(None))
        return self._connection.execute(query).fetchall()

    def archive_dataset(self, dataset_id) -> bool:
        r = self._connection.execute(
            update(Dataset.__table__)  # type: ignore[arg-type]
            .where(Dataset.__table__.c.id == dataset_id)
            .where(Dataset.__table__.c.archived == None)
            .values(archived=func.now())
        )
        return r.rowcount > 0

    def restore_dataset(self, dataset_id) -> bool:
        r = self._connection.execute(
            update(Dataset.__table__)  # type: ignore[arg-type]
            .where(Dataset.__table__.c.id == dataset_id)
            .values(archived=None)
        )
        return r.rowcount > 0

    def delete_dataset(self, dataset_id) -> bool:
        for table in search_field_indexes.values():
            self._connection.execute(
                delete(table).where(table.dataset_ref == dataset_id)
            )
        for crs in self._db.spatially_indexed_crses():
            SpatialIndex = self._db.spatial_index(crs)  # noqa: N806
            self._connection.execute(
                delete(SpatialIndex).where(SpatialIndex.dataset_ref == dataset_id)
            )
        r = self._connection.execute(delete(Dataset).where(Dataset.id == dataset_id))
        return r.rowcount > 0

    def get_dataset(self, dataset_id):
        return self._connection.execute(
            select(*_dataset_select_fields()).where(Dataset.id == dataset_id)
        ).first()

    def get_datasets(self, dataset_ids):
        return self._connection.execute(
            select(*_dataset_select_fields()).where(Dataset.id.in_(dataset_ids))
        ).fetchall()

    def get_derived_datasets(self, dataset_id):
        raise NotImplementedError()

    def get_dataset_sources(self, dataset_id):
        raise NotImplementedError()

    def search_datasets_by_metadata(
        self, metadata: dict, archived: bool | None
    ) -> dict:
        """
        Find any datasets that have the given metadata.
        """
        # Find any storage types whose 'dataset_metadata' document is a subset of the metadata.
        where = Dataset.metadata_doc.contains(metadata)
        if archived:
            where = and_(where, Dataset.archived.is_not(None))
        elif archived is not None:
            where = and_(where, Dataset.archived.is_(None))
        query = select(*_dataset_select_fields()).where(where)
        return self._connection.execute(query).fetchall()

    def search_products_by_metadata(self, metadata: dict) -> dict:
        """
        Find any datasets that have the given metadata.
        """
        # Find any storage types whose 'dataset_metadata' document is a subset of the metadata.
        return self._connection.execute(
            select(Product).where(Product.metadata_doc.contains(metadata))
        ).fetchall()

    @staticmethod
    def _alchemify_expressions(expressions) -> list:
        def raw_expr(expression):
            if isinstance(expression, OrExpression):
                return or_(raw_expr(expr) for expr in expression.exprs)
            return expression.alchemy_expression

        return [raw_expr(expression) for expression in expressions]

    def geospatial_query(self, geom) -> tuple:
        if not geom.crs:
            raise ValueError("Search geometry must have a CRS")
        SpatialIndex = self._db.spatial_index(geom.crs)  # noqa: N806
        if SpatialIndex is None:
            _LOG.info("No spatial index for crs %s - converting to 4326", geom.crs)
            default_crs = CRS("EPSG:4326")
            geom = geom.to_crs(default_crs)
            SpatialIndex = self._db.spatial_index(default_crs)  # noqa: N806
        geom_sql = geom_alchemy(geom)
        _LOG.info("query geometry = %s (%s)", geom.json, geom.crs)
        spatialquery = func.ST_Intersects(SpatialIndex.extent, geom_sql)
        return SpatialIndex, spatialquery

    def search_datasets_query(
        self,
        expressions: Iterable[Expression],
        source_exprs: Iterable[Expression] | None = None,
        select_fields: Iterable[PgField] | None = None,
        with_source_ids: bool = False,
        limit: int | None = None,
        geom: Geometry | None = None,
        archived: bool | None = False,
        order_by: Iterable | None = None,
    ) -> Select:
        # TODO: lineage handling and source search
        assert source_exprs is None
        assert not with_source_ids

        if not select_fields:
            select_fields = _dataset_fields()

        known_fields = _base_known_fields() | {f.name: f for f in select_fields}

        def _ob_exprs(o):
            if isinstance(o, str):
                if known_fields.get(o.lower()) is not None:
                    return known_fields[o.lower()].alchemy_expression
                raise ValueError(f"Cannot order by unknown field {o}")
            elif isinstance(o, PgField):
                return o.alchemy_expression
            else:
                # assume func, clause, or other expression, and leave as-is
                return o

        order_by = [] if order_by is None else [_ob_exprs(o) for o in order_by]

        select_columns = tuple(
            f.alchemy_expression.label(f.name)  # type: ignore[union-attr]
            for f in select_fields
        )
        if geom:
            SpatialIndex, spatialquery = self.geospatial_query(geom)  # noqa: N806
        else:
            spatialquery = None
            SpatialIndex = None  # noqa: N806

        raw_expressions = PostgisDbAPI._alchemify_expressions(expressions)
        join_tables = PostgisDbAPI._join_tables(expressions, select_fields)
        if archived:
            # True: Return archived datasets ONLY
            where_expr = and_(Dataset.archived.is_not(None), *raw_expressions)
        elif archived is not None:
            # False: Return active datasets ONLY
            where_expr = and_(Dataset.archived.is_(None), *raw_expressions)
        else:
            # None: Return BOTH active and archived datasets
            where_expr = and_(*raw_expressions)
        query = select(*select_columns).select_from(Dataset)
        for joins in join_tables:
            query = query.join(*joins)
        if spatialquery is not None:
            where_expr = and_(where_expr, spatialquery)
            query = query.join(SpatialIndex)
        query = query.where(where_expr).order_by(*order_by).limit(limit)
        return query

    def search_datasets(
        self,
        expressions: Iterable[Expression],
        source_exprs: Iterable[Expression] | None = None,
        select_fields: Iterable[PgField] | None = None,
        with_source_ids: bool = False,
        limit: int | None = None,
        geom: Geometry | None = None,
        archived: bool | None = False,
        order_by=None,
    ) -> Iterator:
        """
        :return: An iterable of tuples of decoded values
        """
        assert source_exprs is None
        assert not with_source_ids
        if select_fields is None:
            select_fields = _dataset_fields()
        select_query = self.search_datasets_query(
            expressions,
            select_fields=select_fields,
            limit=limit,
            geom=geom,
            archived=archived,
            order_by=order_by,
        )
        _LOG.debug("search_datasets SQL: %s", str(select_query))

        def decode_row(raw: Iterable[Any]) -> dict[str, Any]:
            return {f.name: f.normalise_value(r) for r, f in zip(raw, select_fields)}

        for row in self._connection.execute(select_query):
            yield decode_row(row)

    def bulk_simple_dataset_search(self, products=None, batch_size: int = 0):
        """
        Perform bulk database reads (e.g. for index cloning)

        Note that this operates with product ids to prevent an unnecessary join to the Product table.

        :param products: Optional iterable of product IDs.  Only fetch nominated products.
        :param batch_size: Number of streamed rows to fetch from database at once.
                           Defaults to zero, which means no streaming.
                           Note streaming is only supported inside a transaction.
        :return: Iterable of tuples of:
                 * Product ID
                 * Dataset metadata document
                 * array of uris
        """
        if batch_size > 0 and not self.in_transaction:
            raise ValueError("Postgresql bulk reads must occur within a transaction.")
        query = (
            select(*_dataset_bulk_select_fields())
            .select_from(Dataset)
            .where(Dataset.archived.is_(None))
        )
        if products:
            query = query.where(Dataset.product_ref.in_(products))

        if batch_size > 0:
            conn = self._connection.execution_options(
                stream_results=True, yield_per=batch_size
            )
        else:
            conn = self._connection
        return conn.execute(query)

    def get_all_lineage(self, batch_size: int):
        """
        Stream all lineage data in bulk (e.g. for index cloning)

        :param batch_size: The number of lineage records to return at once.
        :return: Streamable SQLAlchemy result object.
        """
        if batch_size > 0 and not self.in_transaction:
            raise ValueError("Postgresql bulk reads must occur within a transaction.")
        query = select(
            DatasetLineage.derived_dataset_ref,
            DatasetLineage.classifier,
            DatasetLineage.source_dataset_ref,
        )
        return self._connection.execution_options(
            stream_results=True, yield_per=batch_size
        ).execute(query)

    def insert_lineage_bulk(self, values) -> tuple:
        """
        Insert bulk lineage records (e.g. for index cloning)

        :param values: An array of values dicts for bulk insert
        :return: Tuple[count of rows loaded, count of rows skipped]
        """
        requested = len(values)
        # Simple bulk insert with on_conflict_do_nothing.
        # No need to check referential integrity as this is an external lineage index driver.
        res = self._connection.execute(
            insert(DatasetLineage).on_conflict_do_nothing(), values
        )
        return res.rowcount, requested - res.rowcount

    def get_duplicates(
        self, match_fields: Sequence[PgField], expressions: Sequence[PgExpression]
    ) -> Iterable[dict[str, Any]]:
        # TODO
        if "time" in [f.name for f in match_fields]:
            yield from self.get_duplicates_with_time(match_fields, expressions)

        group_expressions = tuple(f.alchemy_expression for f in match_fields)
        join_tables = PostgisDbAPI._join_tables(expressions, match_fields)

        query = select(
            func.array_agg(Dataset.id).label("ids"), *group_expressions
        ).select_from(Dataset)
        for joins in join_tables:
            query = query.join(*joins)

        query = (
            query.where(
                and_(
                    Dataset.archived.is_(None),
                    *(PostgisDbAPI._alchemify_expressions(expressions)),
                )
            )
            .group_by(*group_expressions)
            .having(func.count(Dataset.id) > 1)
        )
        for row in self._connection.execute(query):
            drow = {"ids": row.ids}
            for f in match_fields:
                drow[f.name] = getattr(row, f.name)
            yield drow

    def get_duplicates_with_time(
        self, match_fields: Sequence[PgField], expressions: Sequence[PgExpression]
    ) -> Iterable[dict[str, Any]]:
        fields = []
        for fld in match_fields:
            if fld.name == "time":
                time_field = type_cast(DateRangeDocField, fld)
            else:
                fields.append(fld.alchemy_expression)

        join_tables = PostgisDbAPI._join_tables(expressions, match_fields)

        cols = [Dataset.id, time_field.expression_with_leniency.label("time"), *fields]
        query = select(*cols).select_from(Dataset)
        for joins in join_tables:
            query = query.join(*joins)

        query = query.where(
            and_(
                Dataset.archived.is_(None),
                *(PostgisDbAPI._alchemify_expressions(expressions)),
            )
        )

        t1 = query.alias("t1")
        t2 = query.alias("t2")

        t1fields = [getattr(t1.c, f.name) for f in fields]  # type: ignore[union-attr]
        time_overlap = (
            select(
                t1.c.id,
                t1.c.time.intersection(t2.c.time).label("time_intersect"),
                *t1fields,
            )
            .select_from(
                t1.join(t2, and_(t1.c.time.overlaps(t2.c.time), t1.c.id != t2.c.id))
            )
            .cte("time_overlap")
        )

        tovlap_fields = [getattr(time_overlap.c, f.name) for f in fields]  # type: ignore[union-attr]
        query = (
            select(
                func.array_agg(func.distinct(time_overlap.c.id)).label("ids"),
                *tovlap_fields,
                text("time_intersect as time"),
            )
            .select_from(time_overlap)
            .group_by(*tovlap_fields, text("time_intersect"))
            .having(func.count(time_overlap.c.id) > 1)
        )

        for row in self._connection.execute(query):
            # TODO: Use decode_rows above - would require creating a field class for the ids array.
            drow: dict[str, Any] = {
                "ids": row.ids,
            }
            for f in fields:
                drow[f.key] = getattr(row, f.key)  # type: ignore[union-attr]
            drow["time"] = time_field.normalise_value((row.time.lower, row.time.upper))
            yield drow

    def count_datasets(
        self,
        expressions: Iterable[Expression],
        archived: bool | None = False,
        geom: Geometry | None = None,
    ) -> int:
        raw_expressions = self._alchemify_expressions(expressions)
        if archived:
            where_expressions = and_(Dataset.archived.is_not(None), *raw_expressions)
        elif archived is not None:
            where_expressions = and_(Dataset.archived.is_(None), *raw_expressions)
        else:
            where_expressions = and_(*raw_expressions)

        query = select(func.count(Dataset.id))

        if geom:
            SpatialIndex, spatialquery = self.geospatial_query(geom)  # noqa: N806
            where_expressions = and_(where_expressions, spatialquery)
            query = query.join(SpatialIndex)

        for join in self._join_tables(expressions=expressions):
            query = query.join(*join)

        select_query = query.where(where_expressions)
        return self._connection.scalar(select_query)

    def count_datasets_through_time(
        self,
        start: datetime.datetime,
        end: datetime.datetime,
        period,
        time_field,
        expressions: Iterable[Expression],
    ) -> Iterator[tuple[tuple[datetime.datetime, datetime.datetime], int]]:
        results = self._connection.execute(
            self.count_datasets_through_time_query(
                start, end, period, time_field, expressions
            )
        )

        for time_period, dataset_count in results:
            # if not time_period.upper_inf:
            yield Range(time_period.lower, time_period.upper), dataset_count

    def count_datasets_through_time_query(
        self, start, end, period, time_field, expressions
    ):
        raw_expressions = self._alchemify_expressions(expressions)

        start_times = select(
            func.generate_series(start, end, cast(period, INTERVAL)).label(
                "start_time"
            ),
        ).alias("start_times")

        time_range_select = (
            select(
                func.tstzrange(
                    start_times.c.start_time, func.lead(start_times.c.start_time).over()
                ).label("time_period"),
            )
        ).alias("all_time_ranges")

        # Exclude the trailing (end time to infinite) row. Is there a simpler way?
        time_ranges = (
            select(
                time_range_select,
            ).where(~func.upper_inf(time_range_select.c.time_period))
        ).alias("time_ranges")

        count_query = select(func.count("*"))
        join_tables = self._join_tables(expressions)
        for joins in join_tables:
            count_query = count_query.join(*joins)
        count_query = count_query.where(
            and_(
                time_field.alchemy_expression.overlaps(time_ranges.c.time_period),
                Dataset.archived == None,
                *raw_expressions,
            )
        )

        return select(time_ranges.c.time_period, count_query.label("dataset_count"))

    def update_search_index(
        self, product_names: Sequence[str] = [], dsids: Sequence[DSID] = []
    ) -> int:
        """
        Update search indexes
        :param product_names: Product names to update
        :param dsids: Dataset IDs to update

        if neither product_names nor dataset ids are supplied, update nothing (N.B. NOT all datasets)

        if both are supplied, both the named products and identified datasets are updated.

        :return:  Number of datasets whose search indexes have been updated.
        """
        if not product_names and not dsids:
            return 0

        ds_query = (
            select(
                Dataset.id,
                Dataset.metadata_doc,
                MetadataType.definition,
            )
            .select_from(Dataset)
            .join(MetadataType)
        )
        if product_names:
            ds_query = ds_query.join(Product)
        if product_names and dsids:
            ds_query = ds_query.where(
                or_(Product.name.in_(product_names), Dataset.id.in_(dsids))
            )
        elif product_names:
            ds_query = ds_query.where(Product.name.in_(product_names))
        elif dsids:
            ds_query = ds_query.where(Dataset.id.in_(dsids))
        rowcount = 0
        for result in self._connection.execute(ds_query):
            dsid, ds_metadata, mdt_def = result
            search_field_vals = extract_dataset_search_fields(ds_metadata, mdt_def)
            for field_name, field_info in search_field_vals.items():
                fld_type, fld_val = field_info
                search_idx = search_field_index_map[fld_type]
                self.insert_dataset_search(search_idx, dsid, field_name, fld_val)
            rowcount += 1
        return rowcount

    def update_spindex(
        self,
        crs_seq: Sequence[CRS] = [],
        product_names: Sequence[str] = [],
        dsids: Sequence[DSID] = [],
    ) -> int:
        """
        Update a spatial index
        :param crs_seq: CRSs for Spatial Indexes to update. Default=all indexes
        :param product_names: Product names to update
        :param dsids: Dataset IDs to update

        if neither product_names nor dataset ids are supplied, update for all datasets.

        if both are supplied, both the named products and identified datasets are updated.

        :return:  Number of spatial index entries updated or verified as unindexed.
        """
        verified = 0
        crses = list(crs_seq) if crs_seq else self._db.spatially_indexed_crses()

        # Update implementation.
        # Design will change, but this method should be fairly low level to be as efficient as possible
        query = select(
            Dataset.id, Dataset.metadata_doc["grid_spatial"]["projection"]
        ).select_from(Dataset)
        if product_names:
            query = query.join(Product)
        if product_names and dsids:
            query = query.where(
                or_(Product.name.in_(product_names), Dataset.id.in_(dsids))
            )
        elif product_names:
            query = query.where(Product.name.in_(product_names))
        elif dsids:
            query = query.where(Dataset.id.in_(dsids))
        for result in self._connection.execute(query):
            dsid = result[0]
            geom = extract_geometry_from_eo3_projection(result[1])
            if not geom:
                verified += 1
                continue
            for crs in crses:
                self.insert_dataset_spatial(dsid, crs, geom)
                verified += 1

        return verified

    @staticmethod
    def _join_tables(expressions=None, fields=None):
        join_args = set()
        if expressions:
            join_args.update(
                expression.field.dataset_join_args for expression in expressions
            )
        if fields:
            join_args.update(field.dataset_join_args for field in fields)
        join_args.discard((Dataset.__table__,))
        # Sort simple joins before qualified joins
        return sorted(join_args, key=len)

    def get_product(self, id_):
        return self._connection.execute(
            select(Product).where(Product.id == id_)
        ).first()

    def get_metadata_type(self, id_: int):
        return self._connection.execute(
            select(MetadataType).where(MetadataType.id == id_)
        ).first()

    def get_product_by_name(self, name: str):
        return self._connection.execute(
            select(Product).where(Product.name == name)
        ).first()

    def get_metadata_type_by_name(self, name: str):
        return self._connection.execute(
            select(MetadataType).where(MetadataType.name == name)
        ).first()

    def insert_product(self, name: str, metadata, metadata_type_id, definition):
        res = self._connection.execute(
            insert(Product.__table__).values(  # type: ignore[arg-type]
                name=name,
                metadata=metadata,
                metadata_type_ref=metadata_type_id,
                definition=definition,
            )
        )

        type_id = res.inserted_primary_key[0]

        return type_id

    def insert_product_bulk(self, values) -> tuple:
        requested = len(values)
        res = self._connection.execute(insert(Product), values)
        return res.rowcount, requested - res.rowcount

    def update_product(
        self,
        name: str,
        metadata,
        metadata_type_id,
        definition,
        update_metadata_type: bool = False,
    ):
        res = self._connection.execute(
            update(Product.__table__)  # type: ignore[arg-type]
            .returning(Product.__table__.c.id)
            .where(Product.__table__.c.name == name)
            .values(
                metadata=metadata,
                metadata_type_ref=metadata_type_id,
                definition=definition,
            )
        )
        prod_id = res.first()[0]

        if update_metadata_type:
            if not self._connection.in_transaction():
                raise RuntimeError("Must update metadata types in transaction")

            self._connection.execute(
                update(Dataset.__table__)  # type: ignore[arg-type]
                .where(Dataset.__table__.c.product_ref == prod_id)
                .values(
                    metadata_type_ref=metadata_type_id,
                )
            )

        return prod_id

    def delete_product(self, name: str):
        res = self._connection.execute(
            delete(Product.__table__)  # type: ignore[arg-type]
            .returning(Product.__table__.c.id)
            .where(Product.__table__.c.name == name)
        )

        return res.first()[0]

    def insert_metadata_type(self, name: str, definition):
        res = self._connection.execute(
            insert(MetadataType.__table__).values(  # type: ignore[arg-type]
                name=name, definition=definition
            )
        )
        return res.inserted_primary_key[0]

    def insert_metadata_bulk(self, values) -> tuple:
        requested = len(values)
        res = self._connection.execute(
            insert(
                MetadataType.__table__  # type: ignore[arg-type]
            ).on_conflict_do_nothing(index_elements=["id"]),
            values,
        )
        return res.rowcount, requested - res.rowcount

    def update_metadata_type(self, name: str, definition):
        res = self._connection.execute(
            update(MetadataType.__table__)  # type: ignore[arg-type]
            .returning(MetadataType.__table__.c.id)
            .where(MetadataType.__table__.c.name == name)
            .values(name=name, definition=definition)
        )
        return res.first()[0]

    @staticmethod
    def _get_active_field_names(fields, metadata_doc) -> Iterator:
        for field in fields.values():
            if field.can_extract:
                try:
                    value = field.extract(metadata_doc)
                    if value is not None:
                        yield field.name
                except (AttributeError, KeyError, ValueError):
                    continue

    def get_all_products(self):
        return self._connection.execute(
            select(Product).order_by(Product.name.asc())
        ).fetchall()

    def get_all_product_docs(self):
        return self._connection.execute(select(Product.definition))

    def _get_products_for_metadata_type(self, id_):
        return self._connection.execute(
            select(Product)
            .where(Product.metadata_type_ref == id_)
            .order_by(Product.name.asc())
        ).fetchall()

    def get_all_metadata_types(self):
        return self._connection.execute(
            select(MetadataType).order_by(MetadataType.name.asc())
        ).fetchall()

    def get_all_metadata_type_defs(self) -> Iterator:
        for r in self._connection.execute(
            select(MetadataType.definition).order_by(MetadataType.name.asc())
        ):
            yield r[0]

    def get_location(self, dataset_id):
        return self._connection.execute(
            select(Dataset.uri).where(Dataset.id == dataset_id)
        ).first()

    def remove_location(self, dataset_id, uri) -> bool:
        """
        Remove a dataset's location

        :returns bool: Was the location deleted?
        """
        scheme, body = split_uri(uri)
        res = self._connection.execute(
            update(Dataset.__table__)  # type: ignore[arg-type]
            .where(
                and_(
                    Dataset.__table__.c.id == dataset_id,
                    Dataset.__table__.c.uri_scheme == scheme,
                    Dataset.__table__.c.uri_body == body,
                )
            )
            .values(uri_scheme=None, uri_body=None)
        )
        return res.rowcount > 0

    @override
    def __repr__(self) -> str:
        return f"PostgresDb<connection={self._connection!r}>"

    def list_users(self) -> Iterator:
        result = self._connection.execute(
            text("""
            select
                group_role.rolname as role_name,
                user_role.rolname as user_name,
                pg_catalog.shobj_description(user_role.oid, 'pg_authid') as description
            from pg_roles group_role
            inner join pg_auth_members am on am.roleid = group_role.oid
            inner join pg_roles user_role on am.member = user_role.oid
            where (group_role.rolname like 'odc_%%') and not (user_role.rolname like 'odc_%%')
            order by group_role.oid asc, user_role.oid asc;
        """)
        )
        for row in result:
            yield _core.from_pg_role(row.role_name), row.user_name, row.description

    def create_user(
        self, username: str, password: str, role, description: str | None = None
    ) -> None:
        pg_role = _core.to_pg_role(role)
        username = escape_pg_identifier(self._connection, username)
        sql = text(f"create user {username} password :password in role {pg_role}")
        self._connection.execute(sql, {"password": password})
        if description:
            sql = text(f"comment on role {username} is :description")
            self._connection.execute(sql, {"description": description})

    def drop_users(self, users: Iterable[str]) -> None:
        for username in users:
            sql = text(f"drop role {escape_pg_identifier(self._connection, username)}")
            self._connection.execute(sql)

    def grant_role(self, role: str, users: Iterable[str]) -> None:
        """
        Grant a role to a user.
        """
        pg_role = _core.to_pg_role(role)

        for user in users:
            if not _core.has_role(self._connection, user):
                raise ValueError(f"Unknown user {user!r}")

        _core.grant_role(self._connection, pg_role, users)

    def insert_home(
        self, home: str, ids: Iterable[uuid.UUID], allow_updates: bool
    ) -> int:
        """
        Set home for multiple IDs (but one home value)

        :param home: The home value to set
        :param ids: The IDs to set it for
        :param allow_updates: If False only inserts are allowed
        :return: number of database records updated or added.
        """
        values = [{"dataset_ref": id_, "home": home} for id_ in ids]
        qry = insert(DatasetHome)
        if allow_updates:
            qry = qry.on_conflict_do_update(
                index_elements=["dataset_ref"],
                set_={"home": home},
                where=(DatasetHome.home != home),
            )
        try:
            res = self._connection.execute(qry, values)
            return res.rowcount
        except IntegrityError:
            return 0

    def delete_home(self, ids):
        """
        Delete the home value for the specified IDs

        :param ids: The IDs to delete home for
        :return: The number of hone records deleted from the databes
        """
        res = self._connection.execute(
            delete(DatasetHome).where(DatasetHome.dataset_ref.in_(ids))
        )
        return res.rowcount

    def select_homes(self, ids) -> dict:
        """
        Find homes for IDs.

        :param ids: Iterable of IDs
        :return: Mapping of ID to home string for IDs found in database.
        """
        results = self._connection.execute(
            select(DatasetHome).where(DatasetHome.dataset_ref.in_(ids))
        )
        return {row.dataset_ref: row.home for row in results}

    def get_all_relations(
        self, dsids: Iterable[uuid.UUID]
    ) -> Iterable[LineageRelation]:
        """
        Fetch all lineage relations in the database involving a set on dataset IDs.

        :param dsids: Iterable of dataset IDs
        :return: Iterable of LineageRelation objects.
        """
        results = self._connection.execute(
            select(DatasetLineage).where(
                or_(
                    DatasetLineage.derived_dataset_ref.in_(dsids),
                    DatasetLineage.source_dataset_ref.in_(dsids),
                )
            )
        )
        for rel in results:
            yield LineageRelation(
                classifier=rel.classifier,
                source_id=rel.source_dataset_ref,
                derived_id=rel.derived_dataset_ref,
            )

    def write_relations(
        self, relations: Iterable[LineageRelation], allow_updates: bool
    ) -> int:
        """
        Write a set of LineageRelation objects to the database.

        :param relations: An Iterable of LineageRelation objects
        :param allow_updates: if False, only allow adding new relations, not updating old ones.
        :return: Count of database rows affected
        """
        affected = 0
        if allow_updates:
            by_classifier: dict[str, Any] = {}
            for rel in relations:
                db_repr = {
                    "derived_dataset_ref": rel.derived_id,
                    "source_dataset_ref": rel.source_id,
                    "classifier": rel.classifier,
                }
                if rel.classifier in by_classifier:
                    by_classifier[rel.classifier].append(db_repr)
                else:
                    by_classifier[rel.classifier] = [db_repr]
                for classifier, values in by_classifier.items():
                    qry = insert(DatasetLineage).on_conflict_do_update(
                        index_elements=["derived_dataset_ref", "source_dataset_ref"],
                        set_={"classifier": classifier},
                        where=(DatasetLineage.classifier != classifier),
                    )
                    res = self._connection.execute(qry, values)
                    affected += res.rowcount
        else:
            for rel in relations:
                values = [
                    {
                        "derived_dataset_ref": rel.derived_id,
                        "source_dataset_ref": rel.source_id,
                        "classifier": rel.classifier,
                    }
                ]
                qry = insert(DatasetLineage)
                try:
                    res = self._connection.execute(qry, values)
                    affected += res.rowcount
                except IntegrityError:
                    return 0
        return affected

    def load_lineage_relations(
        self,
        roots: Iterable[uuid.UUID],
        direction: LineageDirection,
        depth: int,
        ids_so_far: set[uuid.UUID] | None = None,
    ) -> Iterable[LineageRelation]:
        """
        Read from the database all indexed LineageRelation objects required to build all LineageTrees with
        the given roots, direction and depth.

        :param roots: Iterable of root dataset ids
        :param direction: tree direction
        :param depth: Maximum tree depth - zero indicates unlimited depth.
        :param ids_so_far: Used for maintaining state through recursion - expected to be None on initial call
        :return: Iterable of LineageRelation objects read from database
        """
        # Naive manually-recursive initial implementation.
        # TODO: Reimplement using WITH RECURSIVE query
        if ids_so_far is None:
            ids_so_far = set(roots)
        qry = select(DatasetLineage)
        if direction == LineageDirection.SOURCES:
            qry = qry.where(DatasetLineage.derived_dataset_ref.in_(roots))
        else:
            qry = qry.where(DatasetLineage.source_dataset_ref.in_(roots))
        relations = []
        next_lvl_ids = set()
        results = self._connection.execute(qry)
        for row in results:
            rel = LineageRelation(
                classifier=row.classifier,
                source_id=row.source_dataset_ref,
                derived_id=row.derived_dataset_ref,
            )
            relations.append(rel)
            next_id = (
                rel.source_id
                if direction == LineageDirection.SOURCES
                else rel.derived_id
            )
            if next_id not in ids_so_far:
                next_lvl_ids.add(next_id)
                ids_so_far.add(next_id)
        next_depth = depth - 1
        recurse = True
        if depth == 0:
            next_depth = 0
        elif depth == 1:
            recurse = False
        if recurse and next_lvl_ids:
            relations.extend(
                self.load_lineage_relations(
                    next_lvl_ids, direction, next_depth, ids_so_far
                )
            )
        return relations

    def remove_lineage_relations(
        self, ids: Iterable[DSID], direction: LineageDirection
    ) -> int:
        """
        Remove lineage relations from the provided ids in the specified direction.

        Note no depth parameter - depth is effectively always 1.

        :param ids: Iterable of IDs to remove lineage information for.
        :param direction: Remove the source or derived lineage relation records
        :return: Return number of relation records deleted.
        """
        qry = delete(DatasetLineage)
        if direction == LineageDirection.SOURCES:
            qry = qry.where(DatasetLineage.derived_dataset_ref.in_(ids))
        else:
            qry = qry.where(DatasetLineage.source_dataset_ref.in_(ids))
        results = self._connection.execute(qry)
        return results.rowcount

    def temporal_extent_by_prod(
        self, product_id: int
    ) -> tuple[datetime.datetime, datetime.datetime]:
        query = self.temporal_extent_full().where(Dataset.product_ref == product_id)
        res = self._connection.execute(query)
        for tmin, tmax in res:
            return time_min.normalise_value(tmin), time_max.normalise_value(tmax)
        raise RuntimeError("Product has no datasets and therefore no temporal extent")

    def temporal_extent_by_ids(
        self, ids: Iterable[DSID]
    ) -> tuple[datetime.datetime, datetime.datetime]:
        query = self.temporal_extent_full().where(Dataset.id.in_(ids))
        res = self._connection.execute(query)
        for tmin, tmax in res:
            return time_min.normalise_value(tmin), time_max.normalise_value(tmax)
        raise ValueError("no dataset ids provided")

    def temporal_extent_full(self) -> Select:
        # Hardcode eo3 standard time locations - do not use this approach in a legacy index driver.

        return select(
            func.min(time_min.alchemy_expression), func.max(time_max.alchemy_expression)
        )

    def find_most_recent_change(self, product_id: int):
        """
        Find the database-local time of the last dataset that changed for this product.
        """
        return self._connection.execute(
            select(
                func.max(
                    func.greatest(
                        Dataset.added,
                        column("updated"),
                    )
                )
            ).where(Dataset.product_ref == product_id)
        ).scalar()



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2025 ODC Contributors
# SPDX-License-Identifier: Apache-2.0

# We often have one-arg-per column, so these checks aren't so useful.
# pylint: disable=too-many-arguments,too-many-public-methods,too-many-lines

# SQLAlchemy queries require "column == None", not "column is None" due to operator overloading:
# pylint: disable=singleton-comparison

"""
Persistence API implementation for postgres.
"""

import datetime
import logging
import uuid  # noqa: F401
from collections.abc import Iterable, Iterator
from typing import Any
from typing import cast as type_cast

from sqlalchemy import (
    Label,
    Select,
    String,
    and_,
    bindparam,
    cast,
    column,
    delete,
    distinct,
    func,
    literal,
    or_,
    select,
    text,
    values,
)
from sqlalchemy.dialects.postgresql import INTERVAL, JSONB, UUID, insert
from sqlalchemy.engine import Row
from sqlalchemy.exc import IntegrityError
from typing_extensions import override

from datacube.index.abstract import DSID
from datacube.index.exceptions import MissingRecordError
from datacube.index.fields import Expression, Field, OrExpression
from datacube.model import Range
from datacube.utils.uris import split_uri

from . import _core
from . import _dynamic as dynamic
from ._fields import (  # noqa: F401
    DateDocField,
    DateRangeDocField,
    NativeField,
    PgExpression,
    PgField,
    SimpleDocField,
    parse_fields,
)
from ._schema import DATASET, DATASET_LOCATION, DATASET_SOURCE, METADATA_TYPE, PRODUCT
from .sql import escape_pg_identifier

PGCODE_FOREIGN_KEY_VIOLATION = "23503"
_LOG: logging.Logger = logging.getLogger(__name__)


def _dataset_uri_field(table):
    return table.c.uri_scheme + ":" + table.c.uri_body


# Fields for selecting dataset with uris
# Need to alias the table, as queries may join the location table for filtering.
SELECTED_DATASET_LOCATION = DATASET_LOCATION.alias("selected_dataset_location")
# All active URIs, from newest to oldest
_ALL_ACTIVE_URIS = func.array(
    select(_dataset_uri_field(SELECTED_DATASET_LOCATION))
    .where(
        and_(
            SELECTED_DATASET_LOCATION.c.dataset_ref == DATASET.c.id,
            SELECTED_DATASET_LOCATION.c.archived == None,
        )
    )
    .order_by(
        SELECTED_DATASET_LOCATION.c.added.desc(), SELECTED_DATASET_LOCATION.c.id.desc()
    )
    .label("uris")
).label("uris")

_DATASET_SELECT_FIELDS = (
    DATASET,
    _ALL_ACTIVE_URIS,
)

_DATASET_BULK_SELECT_FIELDS = (
    PRODUCT.c.name,
    DATASET.c.metadata,
    _ALL_ACTIVE_URIS,
)


def _base_known_fields() -> dict:
    fields = get_native_fields().copy()
    fields["archived"] = NativeField("archived", "Archived date", DATASET.c.archived)
    fields["uris"] = NativeField(
        "uris",
        "all active uris",
        _ALL_ACTIVE_URIS,
    )
    return fields


def get_native_fields() -> dict[str, PgField]:
    # Native fields (hard-coded into the schema)
    fields: dict[str, PgField] = {
        "id": NativeField("id", "Dataset UUID", DATASET.c.id),
        "indexed_time": NativeField(
            "indexed_time", "When dataset was indexed", DATASET.c.added
        ),
        "indexed_by": NativeField(
            "indexed_by", "User who indexed the dataset", DATASET.c.added_by
        ),
        "product": NativeField("product", "Product name", PRODUCT.c.name),
        "product_id": NativeField(
            "product_id", "ID of a product", DATASET.c.dataset_type_ref
        ),
        "metadata_type": NativeField(
            "metadata_type", "Metadata type name of dataset", METADATA_TYPE.c.name
        ),
        "metadata_type_id": NativeField(
            "metadata_type_id", "ID of a metadata type", DATASET.c.metadata_type_ref
        ),
        "metadata_doc": NativeField(
            "metadata_doc", "Full metadata document", DATASET.c.metadata
        ),
        # Fields that can affect row selection
        # Note that this field is a single uri: selecting it will result in one-result per uri.
        # (ie. duplicate datasets if multiple uris, no dataset if no uris)
        "uri": NativeField(
            "uri",
            "Dataset URI",
            DATASET_LOCATION.c.uri_body,
            alchemy_expression=_dataset_uri_field(DATASET_LOCATION),
            affects_row_selection=True,
        ),
    }
    return fields


def get_dataset_fields(metadata_type_definition: dict[str, Any]) -> dict[str, Any]:
    dataset_section = metadata_type_definition["dataset"]

    fields = get_native_fields()
    # "Fixed fields" (not dynamic: defined in metadata type schema)
    fields.update(
        {
            "creation_time": DateDocField(
                "creation_time",
                "Time when dataset was created (processed)",
                DATASET.c.metadata,
                False,
                offset=dataset_section.get("creation_dt") or ("creation_dt",),
            ),
            "format": SimpleDocField(
                "format",
                "File format (GeoTiff, NetCDF)",
                DATASET.c.metadata,
                False,
                offset=dataset_section.get("format") or ("format", "name"),
            ),
            "label": SimpleDocField(
                "label",
                "Label",
                DATASET.c.metadata,
                False,
                offset=dataset_section.get("label") or ("label",),
            ),
        }
    )

    # noinspection PyTypeChecker
    fields.update(parse_fields(dataset_section["search_fields"], DATASET.c.metadata))
    return fields


class PostgresDbAPI:
    def __init__(self, connection) -> None:
        self._connection = connection
        self._sqla_txn = None

    @property
    def in_transaction(self):
        return self._connection.in_transaction()

    def begin(self) -> None:
        self._connection.execution_options(isolation_level="REPEATABLE READ")
        self._sqla_txn = self._connection.begin()

    def _end_transaction(self) -> None:
        self._sqla_txn = None
        self._connection.execution_options(isolation_level="AUTOCOMMIT")

    def commit(self) -> None:
        self._sqla_txn.commit()  # type: ignore[attr-defined]
        self._end_transaction()

    def rollback(self) -> None:
        self._sqla_txn.rollback()  # type: ignore[attr-defined]
        self._end_transaction()

    def execute(self, command):
        return self._connection.execute(command)

    def insert_dataset(self, metadata_doc, dataset_id, product_id):
        """
        Insert dataset if not already indexed.
        :return: whether it was inserted
        """
        dataset_type_ref = bindparam("dataset_type_ref")
        ret = self._connection.execute(
            insert(DATASET)
            .from_select(
                ["id", "dataset_type_ref", "metadata_type_ref", "metadata"],
                select(
                    bindparam("id"),
                    dataset_type_ref,
                    select(PRODUCT.c.metadata_type_ref)
                    .where(PRODUCT.c.id == dataset_type_ref)
                    .label("metadata_type_ref"),
                    bindparam("metadata", type_=JSONB),
                ),
            )
            .on_conflict_do_nothing(index_elements=["id"]),
            {
                "id": dataset_id,
                "dataset_type_ref": product_id,
                "metadata": metadata_doc,
            },
        )
        return ret.rowcount > 0

    def insert_dataset_bulk(self, values) -> tuple:
        requested = len(values)
        res = self._connection.execute(insert(DATASET), values)
        return res.rowcount, requested - res.rowcount

    def update_dataset(self, metadata_doc, dataset_id, product_id) -> bool:
        """
        Update dataset
        """
        res = self._connection.execute(
            DATASET.update()
            .returning(DATASET.c.id)
            .where(
                and_(
                    DATASET.c.id == dataset_id, DATASET.c.dataset_type_ref == product_id
                )
            )
            .values(metadata=metadata_doc)
        )
        return res.rowcount > 0

    def insert_dataset_location(self, dataset_id: DSID, uri: str) -> bool:
        """
        Add a location to a dataset if it is not already recorded.

        Returns True if success, False if this location already existed
        """
        scheme, body = split_uri(uri)

        r = self._connection.execute(
            insert(DATASET_LOCATION).on_conflict_do_nothing(
                index_elements=["uri_scheme", "uri_body", "dataset_ref"]
            ),
            {
                "dataset_ref": dataset_id,
                "uri_scheme": scheme,
                "uri_body": body,
            },
        )

        return r.rowcount > 0

    def insert_dataset_location_bulk(self, values) -> tuple:
        requested = len(values)
        res = self._connection.execute(insert(DATASET_LOCATION), values)
        return res.rowcount, requested - res.rowcount

    def contains_dataset(self, dataset_id) -> bool:
        return bool(
            self._connection.execute(
                select(DATASET.c.id).where(DATASET.c.id == dataset_id)
            ).fetchone()
        )

    def datasets_intersection(self, dataset_ids) -> list:
        """Compute set intersection: db_dataset_ids & dataset_ids"""
        return [
            r[0]
            for r in self._connection.execute(
                select(DATASET.c.id).where(DATASET.c.id.in_(dataset_ids))
            ).fetchall()
        ]

    def get_datasets_for_location(self, uri, mode: str | None = None):
        scheme, body = split_uri(uri)

        if mode is None:
            mode = "exact" if body.count("#") > 0 else "prefix"

        if mode == "exact":
            body_query = DATASET_LOCATION.c.uri_body == body
        elif mode == "prefix":
            body_query = DATASET_LOCATION.c.uri_body.startswith(body)
        else:
            raise ValueError(f"Unsupported query mode {mode}")

        return self._connection.execute(
            select(*_DATASET_SELECT_FIELDS)
            .select_from(DATASET_LOCATION.join(DATASET))
            .where(and_(DATASET_LOCATION.c.uri_scheme == scheme, body_query))
        ).fetchall()

    def all_dataset_ids(self, archived: bool | None = False):
        query = select(DATASET.c.id).select_from(DATASET)
        if archived:
            query = query.where(DATASET.c.archived.is_not(None))
        elif archived is not None:
            query = query.where(DATASET.c.archived.is_(None))
        return self._connection.execute(query).fetchall()

    def insert_dataset_source(self, classifier, dataset_id, source_dataset_id):
        try:
            r = self._connection.execute(
                insert(DATASET_SOURCE).on_conflict_do_nothing(
                    index_elements=["classifier", "dataset_ref"]
                ),
                {
                    "classifier": classifier,
                    "dataset_ref": dataset_id,
                    "source_dataset_ref": source_dataset_id,
                },
            )
            return r.rowcount > 0
        except IntegrityError as e:
            if e.orig.pgcode == PGCODE_FOREIGN_KEY_VIOLATION:
                raise MissingRecordError(
                    "Referenced source dataset doesn't exist"
                ) from None
            raise

    def archive_dataset(self, dataset_id) -> None:
        self._connection.execute(
            DATASET.update()
            .where(DATASET.c.id == dataset_id)
            .where(DATASET.c.archived == None)
            .values(archived=func.now())
        )

    def restore_dataset(self, dataset_id) -> None:
        self._connection.execute(
            DATASET.update().where(DATASET.c.id == dataset_id).values(archived=None)
        )

    def delete_dataset(self, dataset_id) -> None:
        self._connection.execute(
            DATASET_LOCATION.delete().where(
                DATASET_LOCATION.c.dataset_ref == dataset_id
            )
        )
        self._connection.execute(
            DATASET_SOURCE.delete().where(DATASET_SOURCE.c.dataset_ref == dataset_id)
        )
        self._connection.execute(DATASET.delete().where(DATASET.c.id == dataset_id))

    def get_dataset(self, dataset_id):
        return self._connection.execute(
            select(*_DATASET_SELECT_FIELDS).where(DATASET.c.id == dataset_id)
        ).first()

    def get_datasets(self, dataset_ids):
        return self._connection.execute(
            select(*_DATASET_SELECT_FIELDS).where(DATASET.c.id.in_(dataset_ids))
        ).fetchall()

    def get_derived_datasets(self, dataset_id):
        return self._connection.execute(
            select(*_DATASET_SELECT_FIELDS)
            .select_from(
                DATASET.join(
                    DATASET_SOURCE, DATASET.c.id == DATASET_SOURCE.c.dataset_ref
                )
            )
            .where(DATASET_SOURCE.c.source_dataset_ref == dataset_id)
        ).fetchall()

    def get_dataset_sources(self, dataset_id):
        # recursively build the list of (dataset_ref, source_dataset_ref) pairs starting from dataset_id
        # include (dataset_ref, NULL) [hence the left join]
        sources = (
            select(
                DATASET.c.id.label("dataset_ref"),
                DATASET_SOURCE.c.source_dataset_ref,
                DATASET_SOURCE.c.classifier,
            )
            .select_from(
                DATASET.join(
                    DATASET_SOURCE,
                    DATASET.c.id == DATASET_SOURCE.c.dataset_ref,
                    isouter=True,
                )
            )
            .where(DATASET.c.id == dataset_id)
            .cte(name="sources", recursive=True)
        )

        sources = sources.union_all(
            select(
                sources.c.source_dataset_ref.label("dataset_ref"),
                DATASET_SOURCE.c.source_dataset_ref,
                DATASET_SOURCE.c.classifier,
            )
            .select_from(
                sources.join(
                    DATASET_SOURCE,
                    sources.c.source_dataset_ref == DATASET_SOURCE.c.dataset_ref,
                    isouter=True,
                )
            )
            .where(sources.c.source_dataset_ref != None)
        )

        # turn the list of pairs into adjacency list (dataset_ref, [source_dataset_ref, ...])
        # some source_dataset_ref's will be NULL
        aggd = (
            select(
                sources.c.dataset_ref,
                func.array_agg(sources.c.source_dataset_ref).label("sources"),
                func.array_agg(sources.c.classifier).label("classes"),
            )
            .group_by(sources.c.dataset_ref)
            .alias("aggd")
        )

        # join the adjacency list with datasets table
        select_fields = _DATASET_SELECT_FIELDS + (aggd.c.sources, aggd.c.classes)
        query = select(*select_fields).select_from(
            aggd.join(DATASET, DATASET.c.id == aggd.c.dataset_ref)
        )

        return self._connection.execute(query).fetchall()

    def search_datasets_by_metadata(
        self, metadata: dict, archived: bool | None = False
    ) -> dict:
        """
        Find any datasets that have the given metadata.
        """
        # Find any storage types whose 'dataset_metadata' document is a subset of the metadata.
        where_clause = DATASET.c.metadata.contains(metadata)
        if archived:
            where_clause = and_(where_clause, DATASET.c.archived.is_not(None))
        elif archived is not None:
            where_clause = and_(where_clause, DATASET.c.archived.is_(None))
        query = select(*_DATASET_SELECT_FIELDS).where(where_clause)
        return self._connection.execute(query).fetchall()

    def search_products_by_metadata(self, metadata: dict) -> dict:
        """
        Find any products that have the given metadata.
        """
        # Find any products types whose metadata document contains the passed in metadata
        return self._connection.execute(
            PRODUCT.select().where(PRODUCT.c.metadata.contains(metadata))
        ).fetchall()

    @staticmethod
    def _alchemify_expressions(expressions) -> list:
        def raw_expr(expression):
            if isinstance(expression, OrExpression):
                return or_(raw_expr(expr) for expr in expression.exprs)
            return expression.alchemy_expression

        return [raw_expr(expression) for expression in expressions]

    @staticmethod
    def search_datasets_query(
        expressions: Iterable[Expression],
        source_exprs: Iterable[Expression] | None = None,
        select_fields: Iterable[PgField] | None = None,
        with_source_ids: bool = False,
        limit: int | None = None,
        archived: bool | None = False,
        order_by=None,
    ) -> Select:
        if select_fields:
            # Expand select fields, inserting placeholder columns selections for fields that aren't defined for
            # this product query.
            select_columns = tuple(
                f.alchemy_expression.label(f.name) if f is not None else None
                for f in select_fields
            )
            known_fields = _base_known_fields() | {f.name: f for f in select_fields}
        else:
            select_columns = _DATASET_SELECT_FIELDS
            known_fields = _base_known_fields()

        def _ob_exprs(o):
            if isinstance(o, str):
                if known_fields.get(o.lower()) is not None:
                    return known_fields[o.lower()].alchemy_expression
                raise ValueError(f"Cannot order by unknown field {o}")
            elif isinstance(o, PgField):
                return o.alchemy_expression
            else:
                # assume func, clause, or other expression, and leave as-is
                return o

        order_by = [] if order_by is None else [_ob_exprs(o) for o in order_by]

        if with_source_ids:
            # Include the IDs of source datasets
            select_columns += (
                select(func.array_agg(DATASET_SOURCE.c.source_dataset_ref))
                .select_from(DATASET_SOURCE)
                .where(DATASET_SOURCE.c.dataset_ref == DATASET.c.id)
                .group_by(DATASET_SOURCE.c.dataset_ref)
                .label("dataset_refs"),
            )

        raw_expressions = PostgresDbAPI._alchemify_expressions(expressions)
        from_expression = PostgresDbAPI._from_expression(
            DATASET, expressions, select_fields
        )
        if archived:
            # True: Archived datasets only:
            where_expr = and_(DATASET.c.archived.is_not(None), *raw_expressions)
        elif archived is not None:
            # False/default:  Active datasets only:
            where_expr = and_(DATASET.c.archived.is_(None), *raw_expressions)
        else:
            # None: both active and archived datasets
            where_expr = and_(*raw_expressions)

        if not source_exprs:
            return (
                select(*select_columns)
                .select_from(from_expression)
                .where(where_expr)
                .order_by(*order_by)
                .limit(limit)
            )
        select_fields = select_columns + (
            DATASET_SOURCE.c.source_dataset_ref,
            literal(1).label("distance"),
            DATASET_SOURCE.c.classifier.label("path"),
        )
        base_query = (
            select(*select_fields)
            .select_from(
                from_expression.join(
                    DATASET_SOURCE, DATASET.c.id == DATASET_SOURCE.c.dataset_ref
                )
            )
            .where(where_expr)
        ).cte(name="base_query", recursive=True)

        rq_select_cols = [
            col
            for col in base_query.columns
            if col.name not in ["source_dataset_ref", "distance", "path"]
        ] + [
            DATASET_SOURCE.c.source_dataset_ref,
            (base_query.c.distance + 1).label("distance"),
            (base_query.c.path + "." + DATASET_SOURCE.c.classifier).label("path"),
        ]
        recursive_query = base_query.union_all(
            select(*rq_select_cols).select_from(
                base_query.join(
                    DATASET_SOURCE,
                    base_query.c.source_dataset_ref == DATASET_SOURCE.c.dataset_ref,
                )
            )
        )
        if archived:
            where_expr = and_(
                DATASET.c.archived.is_not(None),
                *PostgresDbAPI._alchemify_expressions(source_exprs),
            )
        elif archived is not None:
            where_expr = and_(
                DATASET.c.archived.is_(None),
                *PostgresDbAPI._alchemify_expressions(source_exprs),
            )
        else:
            where_expr = and_(*PostgresDbAPI._alchemify_expressions(source_exprs))

        return (
            select(
                distinct(recursive_query.c.id),
                *[
                    col
                    for col in recursive_query.columns
                    if col.name not in ["id", "source_dataset_ref", "distance", "path"]
                ],
            )
            .select_from(
                recursive_query.join(
                    DATASET, DATASET.c.id == recursive_query.c.source_dataset_ref
                )
            )
            .where(where_expr)
            .order_by(*order_by)
            .limit(limit)
        )

    def search_datasets(
        self,
        expressions: Iterable[Expression],
        source_exprs: Iterable[Expression] | None = None,
        select_fields: Iterable[PgField] | None = None,
        with_source_ids: bool = False,
        limit: int | None = None,
        archived: bool | None = False,
        order_by=None,
    ):
        select_query = self.search_datasets_query(
            expressions,
            source_exprs,
            select_fields,
            with_source_ids,
            limit,
            archived=archived,
            order_by=order_by,
        )
        return self._connection.execute(select_query)

    def bulk_simple_dataset_search(self, products=None, batch_size: int = 0) -> list:
        """
        Perform bulk database reads (e.g. for index cloning)

        :param products: Optional iterable of product names.  Only fetch nominated products.
        :param batch_size: Number of streamed rows to fetch from database at once.
                           Defaults to zero, which means no streaming.
                           Note streaming is only supported inside a transaction.
        :return: Iterable of tuples of:
                 * Product name
                 * Dataset metadata document
                 * array of uris
        """
        if batch_size > 0 and not self.in_transaction:
            raise ValueError("Postgresql bulk reads must occur within a transaction.")
        if products:
            query = (
                select(PRODUCT.c.id)
                .select_from(PRODUCT)
                .where(PRODUCT.c.name.in_(products))
            )
            products = [row[0] for row in self._connection.execute(query)]
            if not products:
                return []
        else:
            products = None
        query = (
            select(*_DATASET_BULK_SELECT_FIELDS)
            .select_from(DATASET)
            .join(PRODUCT)
            .where(DATASET.c.archived == None)
        )
        if products:
            query = query.where(DATASET.c.dataset_type_ref.in_(products))
        return self._connection.execution_options(
            stream_results=True, yield_per=batch_size
        ).execute(query)

    def get_all_lineage(self, batch_size: int):
        """
        Stream all lineage data in bulk (e.g. for index cloning)

        :param batch_size: The number of lineage records to return at once.
        :return: Streamable SQLAlchemy result object.
        """
        if batch_size > 0 and not self.in_transaction:
            raise ValueError("Postgresql bulk reads must occur within a transaction.")
        query = select(
            DATASET_SOURCE.c.dataset_ref,
            DATASET_SOURCE.c.classifier,
            DATASET_SOURCE.c.source_dataset_ref,
        )
        return self._connection.execution_options(
            stream_results=True, yield_per=batch_size
        ).execute(query)

    def insert_lineage_bulk(self, vals) -> tuple:
        """
        Insert bulk lineage records (e.g. for index cloning)

        :param vals: An array of values dicts for bulk insert
        :return: tuple[count of rows loaded, count of rows skipped]
        """
        requested = len(vals)
        # Wrap values in SQLAlchemy Values object
        sqla_vals = values(
            column("dataset_ref", UUID),
            column("classifier", String),
            column("source_dataset_ref", UUID),
            name="batch_data",
        ).data(vals)
        # Join Values object against the dataset table, via both FK relations to
        # filter out external lineage that cannot be loaded into a legacy lineage index driver
        derived_ds = DATASET.alias("derived")
        source_ds = DATASET
        sel_query = sqla_vals.select().where(
            derived_ds.select()
            .where(derived_ds.c.id == sqla_vals.c.dataset_ref)
            .exists(),
            source_ds.select()
            .where(source_ds.c.id == sqla_vals.c.source_dataset_ref)
            .exists(),
        )
        query = (
            insert(DATASET_SOURCE)
            .from_select(["dataset_ref", "classifier", "source_dataset_ref"], sel_query)
            .on_conflict_do_nothing(index_elements=["classifier", "dataset_ref"])
        )
        res = self._connection.execute(query)
        return res.rowcount, requested - res.rowcount

    def get_duplicates(
        self, match_fields: Iterable[Field], expressions: Iterable[Expression]
    ) -> Iterable[Row]:
        if "time" in [f.name for f in match_fields]:
            return self.get_duplicates_with_time(match_fields, expressions)

        group_expressions = tuple(
            type_cast(PgField, f).alchemy_expression for f in match_fields
        )

        select_query = (
            select(func.array_agg(DATASET.c.id).label("ids"), *group_expressions)
            .select_from(
                PostgresDbAPI._from_expression(DATASET, expressions, match_fields)
            )
            .where(
                and_(
                    DATASET.c.archived == None,
                    *(PostgresDbAPI._alchemify_expressions(expressions)),
                )
            )
            .group_by(*group_expressions)
            .having(func.count(DATASET.c.id) > 1)
        )
        return self._connection.execute(select_query)

    def get_duplicates_with_time(
        self, match_fields: Iterable[Field], expressions: Iterable[Expression]
    ) -> Iterable[Row]:
        """
        If considering time when searching for duplicates, we need to grant some amount of leniency
        in case timestamps are not exactly the same.
        From the set of datasets that are active and have the correct product (candidates),
        find all those whose extended timestamp range overlap (overlapping),
        then group them by the other fields.
        """
        fields = []
        time_field: Label[Any] | None = None
        for f in match_fields:
            if f.name == "time":
                time_field = type_cast(DateRangeDocField, f).expression_with_leniency
            else:
                fields.append(type_cast(PgField, f).alchemy_expression)

        if time_field is None:
            raise Exception("No timme field in duplicates query")

        candidates_table = (
            select(DATASET.c.id, time_field.label("time"), *fields)
            .select_from(
                PostgresDbAPI._from_expression(DATASET, expressions, match_fields)
            )
            .where(
                and_(
                    DATASET.c.archived == None,
                    *(PostgresDbAPI._alchemify_expressions(expressions)),
                )
            )
        )

        t1 = candidates_table.alias("t1")
        t2 = candidates_table.alias("t2")

        fields = [getattr(t1.c, f.name) for f in fields]
        overlapping = (
            select(t1.c.id, text("t1.time * t2.time as time_intersect"), *fields)
            .select_from(
                t1.join(t2, and_(t1.c.time.overlaps(t2.c.time), t1.c.id != t2.c.id))
            )
            .cte("time_overlap")
        )

        fields = [getattr(overlapping.c, f.name) for f in fields]
        final_query = (
            select(
                func.array_agg(func.distinct(overlapping.c.id)).label("ids"),
                *fields,
                text(
                    "(lower(time_intersect) at time zone 'UTC', upper(time_intersect) at time zone 'UTC') as time"
                ),
            )
            .select_from(overlapping)
            .group_by(*fields, text("time_intersect"))
            .having(func.count(overlapping.c.id) > 1)
        )

        return self._connection.execute(final_query)

    def count_datasets(
        self, expressions: Iterable[Expression], archived: bool | None = False
    ) -> int:
        raw_expressions = self._alchemify_expressions(expressions)
        if archived:
            where_exprs = and_(DATASET.c.archived.is_not(None), *raw_expressions)
        elif archived is not None:
            where_exprs = and_(DATASET.c.archived.is_(None), *raw_expressions)
        else:
            where_exprs = and_(*raw_expressions)

        select_query = (
            select(func.count())
            .select_from(self._from_expression(DATASET, expressions))
            .where(where_exprs)
        )

        return self._connection.scalar(select_query)

    def count_datasets_through_time(
        self,
        start: datetime.datetime,
        end: datetime.datetime,
        period: str,
        time_field,
        expressions: Iterable[Expression],
    ) -> Iterator[tuple[tuple[datetime.datetime, datetime.datetime], int]]:
        results = self._connection.execute(
            self.count_datasets_through_time_query(
                start, end, period, time_field, expressions
            )
        )

        for time_period, dataset_count in results:
            # if not time_period.upper_inf:
            yield Range(time_period.lower, time_period.upper), dataset_count

    def count_datasets_through_time_query(
        self, start, end, period, time_field, expressions
    ):
        raw_expressions = self._alchemify_expressions(expressions)

        start_times = select(
            func.generate_series(start, end, cast(period, INTERVAL)).label(
                "start_time"
            ),
        ).alias("start_times")

        time_range_select = (
            select(
                func.tstzrange(
                    start_times.c.start_time, func.lead(start_times.c.start_time).over()
                ).label("time_period"),
            )
        ).alias("all_time_ranges")

        # Exclude the trailing (end time to infinite) row. Is there a simpler way?
        time_ranges = (
            select(
                time_range_select,
            ).where(~func.upper_inf(time_range_select.c.time_period))
        ).alias("time_ranges")

        count_query = (
            select(func.count("*"))
            .select_from(self._from_expression(DATASET, expressions))
            .where(
                and_(
                    time_field.alchemy_expression.overlaps(time_ranges.c.time_period),
                    DATASET.c.archived == None,
                    *raw_expressions,
                )
            )
        )

        return select(time_ranges.c.time_period, count_query.label("dataset_count"))

    @staticmethod
    def _from_expression(source_table, expressions=None, fields=None):
        join_tables = set()
        if expressions:
            join_tables.update(
                expression.field.required_alchemy_table for expression in expressions
            )
        if fields:
            # Ignore placeholder columns
            join_tables.update(
                field.required_alchemy_table for field in fields if field
            )
        join_tables.discard(source_table)

        table_order_hack = [
            DATASET_SOURCE,
            DATASET_LOCATION,
            DATASET,
            PRODUCT,
            METADATA_TYPE,
        ]

        from_expression = source_table
        for table in table_order_hack:
            if table in join_tables:
                from_expression = from_expression.join(table)
        return from_expression

    def get_product(self, id_):
        return self._connection.execute(
            PRODUCT.select().where(PRODUCT.c.id == id_)
        ).first()

    def get_metadata_type(self, id_):
        return self._connection.execute(
            METADATA_TYPE.select().where(METADATA_TYPE.c.id == id_)
        ).first()

    def get_product_by_name(self, name: str):
        return self._connection.execute(
            PRODUCT.select().where(PRODUCT.c.name == name)
        ).first()

    def get_metadata_type_by_name(self, name: str):
        return self._connection.execute(
            METADATA_TYPE.select().where(METADATA_TYPE.c.name == name)
        ).first()

    def insert_product(
        self,
        name: str,
        metadata,
        metadata_type_id,
        search_fields,
        definition,
        concurrently: bool = True,
    ):
        res = self._connection.execute(
            PRODUCT.insert().values(
                name=name,
                metadata=metadata,
                metadata_type_ref=metadata_type_id,
                definition=definition,
            )
        )

        type_id = res.inserted_primary_key[0]

        # Initialise search fields.
        self._setup_product_fields(
            type_id,
            name,
            search_fields,
            definition["metadata"],
            concurrently=concurrently,
        )
        return type_id

    def update_product(
        self,
        name: str,
        metadata,
        metadata_type_id,
        search_fields,
        definition,
        update_metadata_type: bool = False,
        concurrently: bool = False,
    ):
        res = self._connection.execute(
            PRODUCT.update()
            .returning(PRODUCT.c.id)
            .where(PRODUCT.c.name == name)
            .values(
                metadata=metadata,
                metadata_type_ref=metadata_type_id,
                definition=definition,
            )
        )
        type_id = res.first()[0]

        if update_metadata_type:
            if not self._connection.in_transaction():
                raise RuntimeError("Must update metadata types in transaction")

            self._connection.execute(
                DATASET.update()
                .where(DATASET.c.dataset_type_ref == type_id)
                .values(
                    metadata_type_ref=metadata_type_id,
                )
            )

        # Initialise search fields.
        self._setup_product_fields(
            type_id,
            name,
            search_fields,
            definition["metadata"],
            concurrently=concurrently,
            rebuild_view=True,
        )
        return type_id

    def delete_product(self, name: str, fields, definition):
        res = self._connection.execute(
            PRODUCT.delete().returning(PRODUCT.c.id).where(PRODUCT.c.name == name)
        )
        type_id = res.first()[0]

        # Update dynamic fields to remove deleted product fields
        self._setup_product_fields(
            type_id,
            name,
            fields,
            definition["metadata"],
            concurrently=False,
            delete=True,
        )

        return type_id

    def insert_metadata_type(
        self, name: str, definition, concurrently: bool = False
    ) -> None:
        res = self._connection.execute(
            METADATA_TYPE.insert().values(name=name, definition=definition)
        )
        type_id = res.inserted_primary_key[0]

        self._setup_metadata_type_fields(
            type_id, name, definition, concurrently=concurrently
        )

    def update_metadata_type(self, name: str, definition, concurrently: bool = False):
        res = self._connection.execute(
            METADATA_TYPE.update()
            .returning(METADATA_TYPE.c.id)
            .where(METADATA_TYPE.c.name == name)
            .values(name=name, definition=definition)
        )
        type_id = res.first()[0]

        self._setup_metadata_type_fields(
            type_id,
            name,
            definition,
            concurrently=concurrently,
            rebuild_views=True,
        )

        return type_id

    def check_dynamic_fields(
        self,
        concurrently: bool = False,
        rebuild_views: bool = False,
        rebuild_indexes: bool = False,
    ) -> None:
        _LOG.info(
            "Checking dynamic views/indexes. (rebuild views=%s, indexes=%s)",
            rebuild_views,
            rebuild_indexes,
        )

        for metadata_type in self.get_all_metadata_types():
            self._setup_metadata_type_fields(
                metadata_type.id,
                metadata_type.name,
                metadata_type.definition,
                rebuild_indexes=rebuild_indexes,
                rebuild_views=rebuild_views,
                concurrently=concurrently,
            )

    def _setup_metadata_type_fields(
        self,
        id_,
        name: str,
        definition,
        rebuild_indexes: bool = False,
        rebuild_views: bool = False,
        concurrently: bool = True,
    ) -> None:
        # Metadata fields are no longer used (all queries are per-dataset-type): exclude all.
        # This will have the effect of removing any old indexes that still exist.
        fields = get_dataset_fields(definition)
        exclude_fields = tuple(fields)

        dataset_filter = and_(
            DATASET.c.archived == None, DATASET.c.metadata_type_ref == id_
        )
        dynamic.check_dynamic_fields(
            self._connection,
            concurrently,
            dataset_filter,
            exclude_fields,
            fields,
            name,
            rebuild_indexes=rebuild_indexes,
            rebuild_view=rebuild_views,
        )

        for product in self._get_products_for_metadata_type(id_):
            self._setup_product_fields(
                product.id,
                product.name,
                fields,
                product.definition["metadata"],
                rebuild_view=rebuild_views,
                rebuild_indexes=rebuild_indexes,
                concurrently=concurrently,
            )

    def _setup_product_fields(
        self,
        id_,
        name: str,
        fields,
        metadata_doc,
        rebuild_indexes: bool = False,
        rebuild_view: bool = False,
        concurrently: bool = True,
        delete: bool = False,
    ) -> None:
        dataset_filter = and_(
            DATASET.c.archived == None, DATASET.c.dataset_type_ref == id_
        )
        if delete:
            excluded_field_names = tuple(field.name for field in fields.values())
        else:
            excluded_field_names = tuple(
                self._get_active_field_names(fields, metadata_doc)
            )

        dynamic.check_dynamic_fields(
            self._connection,
            concurrently,
            dataset_filter,
            excluded_field_names,
            fields,
            name,
            rebuild_indexes=rebuild_indexes,
            rebuild_view=rebuild_view,
            delete_view=delete,
        )

    @staticmethod
    def _get_active_field_names(fields, metadata_doc) -> Iterator:
        for field in fields.values():
            if field.can_extract:
                try:
                    value = field.extract(metadata_doc)
                    if value is not None:
                        yield field.name
                except (AttributeError, KeyError, ValueError):
                    continue

    def get_all_products(self):
        return self._connection.execute(
            PRODUCT.select().order_by(PRODUCT.c.name.asc())
        ).fetchall()

    def get_all_product_docs(self):
        return self._connection.execute(select(PRODUCT.c.definition))

    def _get_products_for_metadata_type(self, id_):
        return self._connection.execute(
            PRODUCT.select()
            .where(PRODUCT.c.metadata_type_ref == id_)
            .order_by(PRODUCT.c.name.asc())
        ).fetchall()

    def get_all_metadata_types(self):
        return self._connection.execute(
            METADATA_TYPE.select().order_by(METADATA_TYPE.c.name.asc())
        ).fetchall()

    def get_all_metadata_type_docs(self):
        return self._connection.execute(select(METADATA_TYPE.c.definition))

    def get_all_metadata_defs(self):
        return [
            r[0]
            for r in self._connection.execute(
                METADATA_TYPE.select(METADATA_TYPE.c.definition).order_by(
                    METADATA_TYPE.c.name.asc()
                )
            ).fetchall()
        ]

    def temporal_extent_by_product(
        self, product_id: int, min_time_offset, max_time_offset
    ) -> tuple[datetime.datetime, datetime.datetime]:
        time_min = DateDocField(
            "aquisition_time_min",
            "Min of time when dataset was acquired",
            DATASET.c.metadata,
            False,  # is it indexed
            offset=min_time_offset,
            selection="least",
        )

        time_max = DateDocField(
            "aquisition_time_max",
            "Max of time when dataset was acquired",
            DATASET.c.metadata,
            False,  # is it indexed
            offset=max_time_offset,
            selection="greatest",
        )

        res = self._connection.execute(
            select(
                func.min(time_min.alchemy_expression),
                func.max(time_max.alchemy_expression),
            ).where(DATASET.c.dataset_type_ref == product_id)
        ).first()

        if res is None:
            raise RuntimeError(
                "Product has no datasets and therefore no temporal extent"
            )
        return res

    def get_locations(self, dataset_id) -> list:
        return [
            record[0]
            for record in self._connection.execute(
                select(_dataset_uri_field(DATASET_LOCATION))
                .where(
                    and_(
                        DATASET_LOCATION.c.dataset_ref == dataset_id,
                        DATASET_LOCATION.c.archived == None,
                    )
                )
                .order_by(DATASET_LOCATION.c.added.desc(), DATASET_LOCATION.c.id.desc())
            ).fetchall()
        ]

    def get_archived_locations(self, dataset_id) -> list:
        """
        Return a list of uris and archived_times for a dataset
        """
        return [
            (location_uri, archived_time)
            for location_uri, archived_time in self._connection.execute(
                select(
                    _dataset_uri_field(DATASET_LOCATION), DATASET_LOCATION.c.archived
                )
                .where(
                    and_(
                        DATASET_LOCATION.c.dataset_ref == dataset_id,
                        DATASET_LOCATION.c.archived != None,
                    )
                )
                .order_by(DATASET_LOCATION.c.added.desc())
            ).fetchall()
        ]

    def remove_location(self, dataset_id, uri) -> bool:
        """
        Remove the given location for a dataset

        :returns bool: Was the location deleted?
        """
        scheme, body = split_uri(uri)
        res = self._connection.execute(
            delete(DATASET_LOCATION).where(
                and_(
                    DATASET_LOCATION.c.dataset_ref == dataset_id,
                    DATASET_LOCATION.c.uri_scheme == scheme,
                    DATASET_LOCATION.c.uri_body == body,
                )
            )
        )
        return res.rowcount > 0

    def archive_location(self, dataset_id, uri) -> bool:
        scheme, body = split_uri(uri)
        res = self._connection.execute(
            DATASET_LOCATION.update()
            .where(
                and_(
                    DATASET_LOCATION.c.dataset_ref == dataset_id,
                    DATASET_LOCATION.c.uri_scheme == scheme,
                    DATASET_LOCATION.c.uri_body == body,
                    DATASET_LOCATION.c.archived == None,
                )
            )
            .values(archived=func.now())
        )
        return res.rowcount > 0

    def restore_location(self, dataset_id, uri) -> bool:
        scheme, body = split_uri(uri)
        res = self._connection.execute(
            DATASET_LOCATION.update()
            .where(
                and_(
                    DATASET_LOCATION.c.dataset_ref == dataset_id,
                    DATASET_LOCATION.c.uri_scheme == scheme,
                    DATASET_LOCATION.c.uri_body == body,
                    DATASET_LOCATION.c.archived != None,
                )
            )
            .values(archived=None)
        )
        return res.rowcount > 0

    @override
    def __repr__(self) -> str:
        return f"PostgresDb<connection={self._connection!r}>"

    def list_users(self) -> Iterator:
        result = self._connection.execute(
            text("""
            select
                group_role.rolname as role_name,
                user_role.rolname as user_name,
                pg_catalog.shobj_description(user_role.oid, 'pg_authid') as description
            from pg_roles group_role
            inner join pg_auth_members am on am.roleid = group_role.oid
            inner join pg_roles user_role on am.member = user_role.oid
            where (group_role.rolname like 'agdc_%%') and not (user_role.rolname like 'agdc_%%')
            order by group_role.oid asc, user_role.oid asc;
        """)
        )
        for row in result:
            yield _core.from_pg_role(row.role_name), row.user_name, row.description

    def create_user(
        self, username: str, password: str, role, description: str | None = None
    ) -> None:
        pg_role = _core.to_pg_role(role)
        username = escape_pg_identifier(self._connection, username)
        sql = text(f"create user {username} password :password in role {pg_role}")
        self._connection.execute(sql, {"password": password})
        if description:
            sql = text(f"comment on role {username} is :description")
            self._connection.execute(sql, {"description": description})

    def drop_users(self, users: Iterable[str]) -> None:
        for username in users:
            sql = text(f"drop role {escape_pg_identifier(self._connection, username)}")
            self._connection.execute(sql)

    def grant_role(self, role: str, users: Iterable[str]) -> None:
        """
        Grant a role to a user.
        """
        pg_role = _core.to_pg_role(role)

        for user in users:
            if not _core.has_role(self._connection, user):
                raise ValueError(f"Unknown user {user!r}")

        _core.grant_role(self._connection, pg_role, users)

    def find_most_recent_change(self, product_id: int):
        """
        Find the database-local time of the last dataset that changed for this product.
        """
        return self._connection.execute(
            select(
                func.max(
                    func.greatest(
                        DATASET.c.added,
                        column("updated"),
                    )
                )
            ).where(DATASET.c.dataset_type_ref == product_id)
        ).scalar()



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2025 ODC Contributors
# SPDX-License-Identifier: Apache-2.0


from .api import (
    GeneralisedCfg,
    GeneralisedEnv,
    GeneralisedPath,
    GeneralisedRawCfg,
    ODCConfig,
    ODCEnvironment,
)
from .cfg import CfgFormat, find_config, parse_text
from .exceptions import ConfigException
from .opt import (
    IAMAuthenticationOptionHandler,
    IntOptionHandler,
    ODCOptionHandler,
    PostgresURLOptionHandler,
    config_options_for_psql_driver,
    psql_url_from_config,
)
from .utils import ConfigDict, check_valid_env_name, check_valid_option, smells_like_ini

__all__ = [
    "CfgFormat",
    "ConfigDict",
    "ConfigException",
    "GeneralisedCfg",
    "GeneralisedEnv",
    "GeneralisedPath",
    "GeneralisedRawCfg",
    "IAMAuthenticationOptionHandler",
    "IntOptionHandler",
    "ODCConfig",
    "ODCConfig",
    "ODCEnvironment",
    "ODCEnvironment",
    "ODCOptionHandler",
    "PostgresURLOptionHandler",
    "check_valid_env_name",
    "check_valid_option",
    "config_options_for_psql_driver",
    "find_config",
    "parse_text",
    "psql_url_from_config",
    "smells_like_ini",
]



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2025 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
from abc import ABC, abstractmethod
from threading import Lock
from typing import Any

from typing_extensions import override

from datacube.index.exceptions import TransactionException
from datacube.utils.generic import thread_local_cache


class AbstractTransaction(ABC):
    """
    Abstract base class for a Transaction Manager.  All index implementations should extend this base class.

    Thread-local storage and locks ensures one active transaction per index per thread.
    """

    def __init__(self, index_id: str) -> None:
        self._connection: Any = None
        self._tls_id = f"txn-{index_id}"
        self._obj_lock = Lock()
        self._controlling_trans = None

    # Main Transaction API
    def begin(self) -> None:
        """
        Start a new transaction.

        Raises an error if a transaction is already active for this thread.

        Calls implementation-specific _new_connection() method and manages thread local storage and locks.
        """
        with self._obj_lock:
            if self._connection is not None:
                raise ValueError(
                    "Cannot start a new transaction as one is already active"
                )
            self._tls_stash()

    def commit(self) -> None:
        """
        Commit the transaction.

        Raises an error if transaction is not active.

        Calls implementation-specific _commit() method, and manages thread local storage and locks.
        """
        with self._obj_lock:
            if self._connection is None:
                raise ValueError("Cannot commit inactive transaction")
            self._commit()
            self._release_connection()
            self._connection = None
            self._tls_purge()

    def rollback(self) -> None:
        """
        Rollback the transaction.

        Raises an error if transaction is not active.

        Calls implementation-specific _rollback() method, and manages thread local storage and locks.
        """
        with self._obj_lock:
            if self._connection is None:
                raise ValueError("Cannot rollback inactive transaction")
            self._rollback()
            self._release_connection()
            self._connection = None
            self._tls_purge()

    @property
    def active(self) -> bool:
        """
        :return:  True if the transaction is active.
        """
        return self._connection is not None

    # Manage thread-local storage
    def _tls_stash(self) -> None:
        """
        Check TLS is empty, create a new connection and stash it.
        :return:
        """
        stored_val = thread_local_cache(self._tls_id)
        if stored_val is not None:
            # stored_val is outermost transaction in a stack of nested transaction.
            self._controlling_trans = stored_val
            self._connection = stored_val._connection
        else:
            self._connection = self._new_connection()
            thread_local_cache(self._tls_id, purge=True)
            thread_local_cache(self._tls_id, self)

    def _tls_purge(self) -> None:
        thread_local_cache(self._tls_id, purge=True)

    # Commit/Rollback exceptions for Context Manager usage patterns
    def commit_exception(self, errmsg: str) -> TransactionException:
        return TransactionException(errmsg, commit=True)

    def rollback_exception(self, errmsg: str) -> TransactionException:
        return TransactionException(errmsg, commit=False)

    # Context Manager Interface
    def __enter__(self) -> "AbstractTransaction":
        self.begin()
        return self

    def __exit__(self, exc_type, exc_value, traceback) -> bool:
        if not self.active:
            # User has already manually committed or rolled back.
            return True
        if exc_type is not None and issubclass(exc_type, TransactionException):
            # User raised a TransactionException,
            if self._controlling_trans:
                # Nested transaction - reraise TransactionException
                return False
            # Commit or rollback as per exception
            if exc_value.commit:
                self.commit()
            else:
                self.rollback()
            # Tell runtime exception is caught and handled.
            return True
        elif exc_value is not None:
            # Any other exception - reraise.  Rollback if outermost transaction
            if not self._controlling_trans:
                self.rollback()
            # Instruct runtime to rethrow exception
            return False
        else:
            # Exited without exception.  Commit if outermost transaction
            if not self._controlling_trans:
                self.commit()
            return True

    # Internal abstract methods for implementation-specific functionality
    @abstractmethod
    def _new_connection(self) -> Any:
        """
        :return: a new index driver object representing a database connection or equivalent against which transactions
        will be executed.
        """

    @abstractmethod
    def _commit(self) -> None:
        """
        Commit the transaction.
        """

    @abstractmethod
    def _rollback(self) -> None:
        """
        Rollback the transaction.
        """

    @abstractmethod
    def _release_connection(self) -> None:
        """
        Release the connection object stored in self._connection
        """


class UnhandledTransaction(AbstractTransaction):
    # Minimal implementation for index drivers with no transaction handling.
    @override
    def _new_connection(self) -> Any:
        return True

    @override
    def _commit(self) -> None:
        pass

    @override
    def _rollback(self) -> None:
        pass

    @override
    def _release_connection(self) -> None:
        pass



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2025 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
from pathlib import Path

import dask
import dask.delayed
import moto
import pytest

from datacube.utils.aws import (
    s3_client,
    s3_fetch,
    s3_url_parse,
)
from datacube.utils.dask import (
    _save_blob_to_file,
    _save_blob_to_s3,
    compute_memory_per_worker,
    compute_tasks,
    get_total_available_memory,
    partition_map,
    pmap,
    save_blob_to_file,
    save_blob_to_s3,
    start_local_dask,
)
from datacube.utils.io import slurp


def test_compute_tasks() -> None:
    try:
        client = start_local_dask(threads_per_worker=1, dashboard_address=None)

        tasks = (dask.delayed(x) for x in range(100))
        xx = list(compute_tasks(tasks, client))
        assert xx == list(range(100))
    finally:
        client.close()
        del client


def test_start_local_dask_dashboard_link(monkeypatch) -> None:
    monkeypatch.setenv("JUPYTERHUB_SERVICE_PREFIX", "user/test/")
    try:
        client = start_local_dask()
        assert client.dashboard_link.startswith("user/test/proxy/")
    finally:
        client.close()
        del client


def test_partition_map() -> None:
    tasks = partition_map(10, str, range(101))
    tt = list(tasks)
    assert len(tt) == 11
    lump = tt[0].compute()
    assert len(lump) == 10
    assert lump == [str(x) for x in range(10)]

    lump = tt[-1].compute()
    assert len(lump) == 1


def test_pmap() -> None:
    try:
        client = start_local_dask(threads_per_worker=1, dashboard_address=None)

        xx_it = pmap(str, range(101), client=client)
        xx = list(xx_it)

        assert xx == [str(x) for x in range(101)]
    finally:
        client.close()
        del client


@pytest.mark.parametrize(
    "blob",
    [
        "some utf8 string",
        b"raw bytes",
    ],
)
def test_save_blob_file_direct(tmpdir, blob) -> None:
    tmpdir = Path(str(tmpdir))
    fname = str(tmpdir / "file.txt")
    mode = "rt" if isinstance(blob, str) else "rb"

    assert _save_blob_to_file(blob, fname) == (fname, True)
    assert slurp(fname, mode=mode) == blob

    fname = str(tmpdir / "missing" / "file.txt")
    assert _save_blob_to_file(blob, fname) == (fname, False)


@pytest.mark.parametrize(
    "blob",
    [
        "some utf8 string",
        b"raw bytes",
    ],
)
def test_save_blob_file(tmpdir, blob, dask_client) -> None:
    tmpdir = Path(str(tmpdir))
    fname = str(tmpdir / "file.txt")
    dask_blob = dask.delayed(blob)
    mode = "rt" if isinstance(blob, str) else "rb"

    rr = save_blob_to_file(dask_blob, fname)
    assert dask_client.compute(rr).result() == (fname, True)
    assert slurp(fname, mode=mode) == blob

    fname = str(tmpdir / "missing" / "file.txt")
    rr = save_blob_to_file(dask_blob, fname)
    assert dask_client.compute(rr).result() == (fname, False)


@pytest.mark.parametrize(
    "blob",
    [
        "some utf8 string",
        b"raw bytes",
    ],
)
def test_save_blob_s3_direct(blob, monkeypatch) -> None:
    region_name = "us-west-2"
    blob2 = blob + blob

    url = "s3://bucket/file.txt"
    url2 = "s3://bucket/file-2.txt"

    bucket, _ = s3_url_parse(url)
    monkeypatch.setenv("AWS_ACCESS_KEY_ID", "fake-key-id")
    monkeypatch.setenv("AWS_SECRET_ACCESS_KEY", "fake-secret")

    with moto.mock_aws():
        s3 = s3_client(region_name=region_name)
        s3.create_bucket(  # type: ignore[attr-defined]
            Bucket=bucket, CreateBucketConfiguration={"LocationConstraint": region_name}
        )

        assert _save_blob_to_s3(blob, url, region_name=region_name) == (url, True)
        assert _save_blob_to_s3(blob2, url2, region_name=region_name) == (url2, True)

        bb1_fetched = s3_fetch(url, s3=s3)
        bb2_fetched = s3_fetch(url2, s3=s3)
        bb1 = bb1_fetched.decode("utf8") if isinstance(blob, str) else bb1_fetched
        bb2 = bb2_fetched.decode("utf8") if isinstance(blob, str) else bb2_fetched
        assert bb1 == blob
        assert bb2 == blob2

        assert _save_blob_to_s3("", "s3://not-a-bucket/f.txt") == (
            "s3://not-a-bucket/f.txt",
            False,
        )


@pytest.mark.parametrize(
    "blob",
    [
        "some utf8 string",
        b"raw bytes",
    ],
)
def test_save_blob_s3(blob, monkeypatch, dask_client) -> None:
    region_name = "us-west-2"

    blob2 = blob + blob

    dask_blob = dask.delayed(blob)
    dask_blob2 = dask.delayed(blob2)

    url = "s3://bucket/file.txt"
    url2 = "s3://bucket/file-2.txt"

    bucket, _ = s3_url_parse(url)
    monkeypatch.setenv("AWS_ACCESS_KEY_ID", "fake-key-id")
    monkeypatch.setenv("AWS_SECRET_ACCESS_KEY", "fake-secret")

    with moto.mock_aws():
        s3 = s3_client(region_name=region_name)
        s3.create_bucket(  # type: ignore[attr-defined]
            Bucket=bucket, CreateBucketConfiguration={"LocationConstraint": region_name}
        )

        rr = save_blob_to_s3(dask_blob, url, region_name=region_name)
        assert rr.compute() == (url, True)

        rr = save_blob_to_s3(dask_blob2, url2, region_name=region_name)
        assert dask_client.compute(rr).result() == (url2, True)

        bb1_fetched = s3_fetch(url, s3=s3)
        bb2_fetched = s3_fetch(url2, s3=s3)
        bb1 = bb1_fetched.decode("utf8") if isinstance(blob, str) else bb1_fetched
        bb2 = bb2_fetched.decode("utf8") if isinstance(blob, str) else bb2_fetched

        assert bb1 == blob
        assert bb2 == blob2


def test_memory_functions(monkeypatch) -> None:
    gig = 10**9

    total_mem = get_total_available_memory()
    default_safety = min(500 * (1 << 20), total_mem // 2)

    assert total_mem - compute_memory_per_worker() == default_safety
    assert total_mem - compute_memory_per_worker(2) * 2 == default_safety

    assert compute_memory_per_worker(mem_safety_margin=1) == total_mem - 1
    assert compute_memory_per_worker(memory_limit="4G") == 4 * gig
    assert compute_memory_per_worker(2, memory_limit="4G") == 2 * gig
    assert (
        compute_memory_per_worker(memory_limit="4G", mem_safety_margin="1G") == 3 * gig
    )

    total_mem = 1 * gig
    monkeypatch.setenv("MEM_LIMIT", str(total_mem))
    assert get_total_available_memory() == 1 * gig
    assert compute_memory_per_worker(mem_safety_margin=1) == total_mem - 1



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2025 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
import os
from uuid import uuid4 as random_uuid

import pytest

from datacube.model import InconsistentLineageException, LineageDirection, LineageTree
from datacube.model.lineage import LineageIDPair, LineageRelations
from datacube.utils import read_documents


def test_directions() -> None:
    assert LineageDirection.SOURCES != LineageDirection.DERIVED
    assert LineageDirection.SOURCES.opposite() == LineageDirection.DERIVED
    assert LineageDirection.DERIVED.opposite() == LineageDirection.SOURCES


def test_ltree_clsmethods(data_folder) -> None:
    root = random_uuid()
    # Minimal tree - root node only
    minimal = LineageTree.from_data(dsid=root)
    assert minimal.dataset_id == root
    assert minimal.direction == LineageDirection.SOURCES
    assert minimal.children is None
    # Check optional args to from_eo3_doc are set correctly.
    optional_args = LineageTree.from_data(
        dsid=root, sources={}, direction=LineageDirection.DERIVED, home="notused"
    )
    assert optional_args == LineageTree(
        direction=LineageDirection.DERIVED, dataset_id=root, children={}, home=None
    )  # Note home is not written to the root node
    doc = next(iter(read_documents(str(os.path.join(data_folder, "ds_eo3.yml")))))[1]
    tree = LineageTree.from_eo3_doc(doc, home="src_home", home_derived="der_home")
    assert tree.home == "der_home"
    assert tree.children is not None
    for child in tree.children["bc"]:
        assert child.home == "src_home"


@pytest.fixture
def shared_tree_ids():
    return {
        "ard1": random_uuid(),
        "l1_1": random_uuid(),
        "l1_2": random_uuid(),
        "l1_3": random_uuid(),
        "atmos": random_uuid(),
    }


@pytest.fixture
def src_tree_ids(shared_tree_ids):
    return {
        "root": random_uuid(),
        "ard1": shared_tree_ids["ard1"],
        "l1_1": shared_tree_ids["l1_1"],
        "l1_2": shared_tree_ids["l1_2"],
        "l1_3": shared_tree_ids["l1_3"],
        "atmos": shared_tree_ids["atmos"],
    }


@pytest.fixture
def src_lineage_tree(src_tree_ids):
    ids = src_tree_ids
    direction = LineageDirection.SOURCES
    return LineageTree(
        dataset_id=ids["root"],
        direction=direction,
        children={
            "ard": [
                LineageTree(
                    dataset_id=ids["ard1"],
                    direction=direction,
                    children={
                        "l1": [
                            LineageTree(
                                dataset_id=ids["l1_1"],
                                direction=direction,
                                home="level1db",
                                children={},
                            ),
                            LineageTree(
                                dataset_id=ids["l1_2"],
                                direction=direction,
                                home="level1db",
                                children={},
                            ),
                            LineageTree(
                                dataset_id=ids["l1_3"],
                                direction=direction,
                                home="level1db",
                                children={},
                            ),
                        ],
                        "atmos_corr": [
                            LineageTree(
                                dataset_id=ids["atmos"],
                                direction=direction,
                                home="level1db",
                                children=None,
                            )
                        ],
                    },
                ),
            ]
        },
    )


def test_lineage_serialisation(src_lineage_tree, src_tree_ids) -> None:
    ids = src_tree_ids
    serialised = src_lineage_tree.serialise()
    assert serialised == {
        "id": str(ids["root"]),
        "sources": {
            "ard": [
                {
                    "id": str(ids["ard1"]),
                    "sources": {
                        "l1": [
                            {"id": str(ids["l1_1"]), "home": "level1db"},
                            {"id": str(ids["l1_2"]), "home": "level1db"},
                            {"id": str(ids["l1_3"]), "home": "level1db"},
                        ],
                        "atmos_corr": [{"id": str(ids["atmos"]), "home": "level1db"}],
                    },
                }
            ]
        },
    }
    tree_out = LineageTree.deserialise(serialised)
    assert tree_out == src_lineage_tree


@pytest.fixture
def src_lineage_tree_diffhome(src_tree_ids):
    ids = src_tree_ids
    direction = LineageDirection.SOURCES
    return LineageTree(
        dataset_id=ids["root"],
        direction=direction,
        children={
            "ard": [
                LineageTree(
                    dataset_id=ids["ard1"],
                    direction=direction,
                    children={
                        "l1": [
                            LineageTree(
                                dataset_id=ids["l1_1"],
                                direction=direction,
                                home="elsewhere",
                                children={},
                            ),
                            LineageTree(
                                dataset_id=ids["l1_2"],
                                direction=direction,
                                home="elsewhere",
                                children={},
                            ),
                            LineageTree(
                                dataset_id=ids["l1_3"], direction=direction, children={}
                            ),
                        ],
                        "atmos_corr": [
                            LineageTree(
                                dataset_id=ids["atmos"],
                                direction=direction,
                                children=None,
                            )
                        ],
                    },
                ),
            ]
        },
    )


@pytest.fixture
def mixed_dir_lineage_tree(src_tree_ids):
    ids = src_tree_ids
    direction = LineageDirection.SOURCES
    return LineageTree(
        dataset_id=ids["root"],
        direction=direction,
        children={
            "ard": [
                LineageTree(
                    dataset_id=ids["ard1"],
                    direction=LineageDirection.DERIVED,
                    children={
                        "l1": [
                            LineageTree(
                                dataset_id=ids["l1_1"], direction=direction, children={}
                            ),
                            LineageTree(
                                dataset_id=ids["l1_2"], direction=direction, children={}
                            ),
                            LineageTree(
                                dataset_id=ids["l1_3"], direction=direction, children={}
                            ),
                        ],
                        "atmos_corr": [
                            LineageTree(
                                dataset_id=ids["atmos"],
                                direction=direction,
                                children=None,
                            )
                        ],
                    },
                ),
            ]
        },
    )


@pytest.fixture
def big_src_tree_ids(shared_tree_ids):
    ids = shared_tree_ids
    return {
        "root": random_uuid(),
        "ard1": ids["ard1"],
        "ard2": random_uuid(),
        "l1_1": ids["l1_1"],
        "l1_2": ids["l1_2"],
        "l1_3": ids["l1_3"],
        "l1_4": random_uuid(),
        "l1_5": random_uuid(),
        "l1_6": random_uuid(),
        "atmos": ids["atmos"],
        "atmos_parent": random_uuid(),
    }


@pytest.fixture
def big_src_lineage_tree(big_src_tree_ids):
    ids = big_src_tree_ids
    direction = LineageDirection.SOURCES
    return LineageTree(
        dataset_id=ids["root"],
        direction=direction,
        children={
            "ard": [
                LineageTree(
                    dataset_id=ids["ard1"],
                    direction=direction,
                    children={
                        "l1": [
                            LineageTree(
                                dataset_id=ids["l1_1"], direction=direction, children={}
                            ),
                            LineageTree(
                                dataset_id=ids["l1_2"], direction=direction, children={}
                            ),
                            LineageTree(
                                dataset_id=ids["l1_3"], direction=direction, children={}
                            ),
                        ],
                        "atmos_corr": [
                            LineageTree(
                                dataset_id=ids["atmos"],
                                direction=direction,
                                children=None,
                            )
                        ],
                    },
                ),
                LineageTree(
                    dataset_id=ids["ard2"],
                    direction=direction,
                    children={
                        "l1": [
                            LineageTree(
                                dataset_id=ids["l1_4"], direction=direction, children={}
                            ),
                            LineageTree(
                                dataset_id=ids["l1_5"], direction=direction, children={}
                            ),
                            LineageTree(
                                dataset_id=ids["l1_6"], direction=direction, children={}
                            ),
                        ],
                        "atmos_corr": [
                            LineageTree(
                                dataset_id=ids["atmos"],
                                direction=direction,
                                children={
                                    "preatmos": [
                                        LineageTree(
                                            dataset_id=ids["atmos_parent"],
                                            direction=direction,
                                            children={},
                                        )
                                    ]
                                },
                            )
                        ],
                    },
                ),
            ]
        },
    )


@pytest.fixture
def classifier_mismatch(big_src_tree_ids):
    ids = big_src_tree_ids
    direction = LineageDirection.SOURCES
    return LineageTree(
        dataset_id=ids["root"],
        direction=direction,
        children={
            "ard": [
                LineageTree(
                    dataset_id=ids["ard1"],
                    direction=direction,
                    children={
                        "l1": [
                            LineageTree(
                                dataset_id=ids["l1_1"], direction=direction, children={}
                            ),
                            LineageTree(
                                dataset_id=ids["l1_2"], direction=direction, children={}
                            ),
                            LineageTree(
                                dataset_id=ids["l1_3"], direction=direction, children={}
                            ),
                        ],
                        "atmos_corr": [
                            LineageTree(
                                dataset_id=ids["atmos"],
                                direction=direction,
                                children=None,
                            )
                        ],
                    },
                ),
                LineageTree(
                    dataset_id=ids["ard2"],
                    direction=direction,
                    children={
                        "lvl1": [
                            LineageTree(
                                dataset_id=ids["l1_4"], direction=direction, children={}
                            ),
                            LineageTree(
                                dataset_id=ids["l1_5"], direction=direction, children={}
                            ),
                            LineageTree(
                                dataset_id=ids["l1_6"], direction=direction, children={}
                            ),
                        ],
                        "atmos_corr": [
                            LineageTree(
                                dataset_id=ids["atmos"],
                                direction=direction,
                                children={
                                    "preatmos": [
                                        LineageTree(
                                            dataset_id=ids["atmos_parent"],
                                            direction=direction,
                                            children={},
                                        )
                                    ]
                                },
                            )
                        ],
                    },
                ),
            ]
        },
    )


@pytest.fixture
def src_lineage_tree_with_bad_diamond(big_src_tree_ids):
    ids = big_src_tree_ids
    direction = LineageDirection.SOURCES
    # This is a test tree with a malformed diamond relationship.
    return LineageTree(
        dataset_id=ids["root"],
        direction=direction,
        children={
            "ard": [
                LineageTree(
                    dataset_id=ids["ard1"],
                    direction=direction,
                    children={
                        "l1": [
                            LineageTree(
                                dataset_id=ids["l1_1"], direction=direction, children={}
                            ),
                            LineageTree(
                                dataset_id=ids["l1_2"], direction=direction, children={}
                            ),
                            LineageTree(
                                dataset_id=ids["l1_3"], direction=direction, children={}
                            ),
                        ],
                        "atmos_corr": [
                            # First half of the diamond relationship:
                            # ard1 depends on atmos (which depends on atmos_parent)
                            LineageTree(
                                dataset_id=ids["atmos"],
                                direction=direction,
                                children={
                                    "preatmos": [
                                        LineageTree(
                                            dataset_id=ids["atmos_parent"],
                                            direction=direction,
                                            children={},
                                        )
                                    ]
                                },
                            )
                        ],
                    },
                ),
                LineageTree(
                    dataset_id=ids["ard2"],
                    direction=direction,
                    children={
                        "l1": [
                            LineageTree(
                                dataset_id=ids["l1_4"], direction=direction, children={}
                            ),
                            LineageTree(
                                dataset_id=ids["l1_5"], direction=direction, children={}
                            ),
                            LineageTree(
                                dataset_id=ids["l1_6"], direction=direction, children={}
                            ),
                        ],
                        # Second half of the diamond relationship:
                        # ard2 also depends on atmos (which depends on atmos_parent)
                        # But the atmos->atmos_parent relationship was already recorded
                        # above, so "children" here should be None or {}.
                        "atmos_corr": [
                            LineageTree(
                                dataset_id=ids["atmos"],
                                direction=direction,
                                children={
                                    "preatmos": [
                                        LineageTree(
                                            dataset_id=ids["atmos_parent"],
                                            direction=direction,
                                            children={},
                                        )
                                    ]
                                },
                            )
                        ],
                    },
                ),
            ]
        },
    )


def test_child_datasets(big_src_lineage_tree, big_src_tree_ids) -> None:
    cds = big_src_lineage_tree.child_datasets()
    for dsid in big_src_tree_ids.values():
        assert dsid == big_src_lineage_tree.dataset_id or dsid in cds


def test_lin_rels_lin_tree_conversions(big_src_lineage_tree, big_src_tree_ids) -> None:
    # Create LRS from LT
    rels1 = LineageRelations(tree=big_src_lineage_tree)
    # Extract LT from LRS
    extracted_tree = rels1.extract_tree(
        big_src_tree_ids["root"], big_src_lineage_tree.direction
    )
    # Confirm extract LT produces same LRS as original LT
    rels2 = LineageRelations(tree=extracted_tree)
    for rel in rels1.relations:
        assert rel in rels2.relations
    # Extract reverse direction tree
    derived_tree = rels1.extract_tree(
        root=big_src_tree_ids["atmos"], direction=LineageDirection.DERIVED
    )
    rels3 = LineageRelations(tree=derived_tree)
    extracted_tree = rels3.extract_tree(
        big_src_tree_ids["root"], direction=big_src_lineage_tree.direction
    )
    rels4 = LineageRelations(tree=extracted_tree)
    for rel in rels4.relations:
        assert rel in rels1.relations


def test_detect_cyclic_deps(big_src_lineage_tree, big_src_tree_ids) -> None:
    # Confirm trivial cyclic dependencies are detected.
    repeated_uuid = random_uuid()
    looped_tree = LineageTree(
        dataset_id=repeated_uuid,
        direction=LineageDirection.SOURCES,
        children={
            "cyclic_self": [
                LineageTree(
                    dataset_id=repeated_uuid,
                    direction=LineageDirection.SOURCES,
                    children={},
                )
            ]
        },
    )
    with pytest.raises(InconsistentLineageException) as e:
        LineageRelations(tree=looped_tree)
    assert "LineageTrees must be acyclic" in str(e.value)
    # Confirm more complex cyclic dependencies can be detected.
    rels = LineageRelations(tree=big_src_lineage_tree)
    breaking_tree = LineageTree(
        dataset_id=big_src_tree_ids["l1_1"],
        direction=LineageDirection.SOURCES,
        children={
            "cyclic_dep": [
                LineageTree(
                    dataset_id=big_src_tree_ids["root"],
                    direction=LineageDirection.SOURCES,
                )
            ]
        },
    )
    with pytest.raises(InconsistentLineageException) as e:
        rels.merge_tree(breaking_tree)
    assert "LineageTrees must be acyclic" in str(e.value)


def test_subtree(
    big_src_lineage_tree, big_src_tree_ids, src_lineage_tree, src_tree_ids
) -> None:
    sub = big_src_lineage_tree.find_subtree(big_src_tree_ids["root"])
    assert sub == big_src_lineage_tree

    # Test multiple nodes, one with children
    sub = big_src_lineage_tree.find_subtree(big_src_tree_ids["atmos"])
    assert sub.dataset_id == big_src_tree_ids["atmos"]
    assert sub.children is not None

    # Test no nodes with children
    sub = src_lineage_tree.find_subtree(src_tree_ids["atmos"])
    assert sub.dataset_id == src_tree_ids["atmos"]
    assert sub.children is None


def test_good_consistency_check(
    big_src_lineage_tree, src_lineage_tree, big_src_tree_ids
) -> None:
    rels1 = LineageRelations(tree=src_lineage_tree)
    rels2 = LineageRelations(tree=big_src_lineage_tree)
    diff = rels1.relations_diff(rels2)
    assert diff[1] == {} and diff[3] == {}
    assert (
        LineageIDPair(
            derived_id=src_lineage_tree.dataset_id, source_id=big_src_tree_ids["ard1"]
        )
        in diff[0]
    )
    diff = rels1.relations_diff(rels2, allow_updates=True)
    assert diff[1] == {} and diff[3] == {}
    assert (
        LineageIDPair(
            derived_id=src_lineage_tree.dataset_id, source_id=big_src_tree_ids["ard1"]
        )
        in diff[0]
    )
    diff = rels1.relations_diff()
    assert diff[1] == {} and diff[3] == {}
    assert (
        LineageIDPair(
            derived_id=src_lineage_tree.dataset_id, source_id=big_src_tree_ids["ard1"]
        )
        in diff[0]
    )


def test_bad_diamond(src_lineage_tree_with_bad_diamond, big_src_tree_ids) -> None:
    # Test detection of trees that have a diamond relationship in which both paths are extended.
    # E.g. If A->B->C->D  and A->E->C->D, the C->D relationship should only be recorded
    # under one branch (B or A) and the other occurrence of C should have no children recorded.
    with pytest.raises(
        InconsistentLineageException, match="Duplicate nodes in LineageTree"
    ):
        LineageRelations(tree=src_lineage_tree_with_bad_diamond)


def test_home_mismatch(big_src_lineage_tree) -> None:
    tree = big_src_lineage_tree
    tree.children["ard"][0].children["atmos_corr"][0].home = "bungalow"
    tree.children["ard"][1].children["atmos_corr"][0].home = "apartment"
    with pytest.raises(
        InconsistentLineageException, match="Tree contains inconsistent homes"
    ):
        LineageRelations(tree=tree)


def test_classifier_mismatch(big_src_lineage_tree, classifier_mismatch) -> None:
    rels1 = LineageRelations(tree=big_src_lineage_tree)
    rels2 = LineageRelations(tree=classifier_mismatch)
    with pytest.raises(
        InconsistentLineageException,
        match="Dataset .* is derived from .* with inconsistent classifiers.",
    ):
        rels1.merge(rels2)


def test_classifier_update(big_src_lineage_tree, classifier_mismatch) -> None:
    rels1 = LineageRelations(tree=big_src_lineage_tree)
    rels2 = LineageRelations(tree=classifier_mismatch)
    diffs = rels1.relations_diff(existing_relations=rels2, allow_updates=True)
    assert len(diffs[1]) > 0


def test_home_update(src_lineage_tree, src_lineage_tree_diffhome) -> None:
    rels1 = LineageRelations(tree=src_lineage_tree)
    rels2 = LineageRelations(tree=src_lineage_tree_diffhome)
    diffs = rels1.relations_diff(existing_relations=rels2, allow_updates=True)
    assert len(diffs[3]) > 0


def test_mixed_dirs(mixed_dir_lineage_tree) -> None:
    with pytest.raises(
        InconsistentLineageException,
        match="Tree contains both derived and source nodes",
    ):
        LineageRelations(tree=mixed_dir_lineage_tree)


def test_merge_tree_limited_depth(big_src_lineage_tree, big_src_tree_ids) -> None:
    ids = big_src_tree_ids
    rels = LineageRelations(tree=big_src_lineage_tree, max_depth=1)
    assert ids["root"] in rels.dataset_ids
    assert ids["ard1"] in rels.dataset_ids
    assert ids["l1_1"] not in rels.dataset_ids
    assert ids["atmos"] not in rels.dataset_ids
    assert ids["atmos_parent"] not in rels.dataset_ids

    rels = LineageRelations(tree=big_src_lineage_tree, max_depth=2)
    assert ids["root"] in rels.dataset_ids
    assert ids["ard1"] in rels.dataset_ids
    assert ids["l1_1"] in rels.dataset_ids
    assert ids["atmos"] in rels.dataset_ids
    assert ids["atmos_parent"] not in rels.dataset_ids

    rels = LineageRelations(tree=big_src_lineage_tree, max_depth=3)
    assert ids["root"] in rels.dataset_ids
    assert ids["ard1"] in rels.dataset_ids
    assert ids["l1_1"] in rels.dataset_ids
    assert ids["atmos"] in rels.dataset_ids
    assert ids["atmos_parent"] in rels.dataset_ids

    rels = LineageRelations(tree=big_src_lineage_tree, max_depth=7)
    assert ids["atmos_parent"] in rels.dataset_ids



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2025 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
import functools
import itertools
import logging
import os
import pickle
import re
import time
from collections.abc import Generator, Sequence
from pathlib import Path

import click
import pandas as pd

from datacube.index import Index
from datacube.ui import click as dc_ui
from datacube.utils import read_documents

_LOG: logging.Logger = logging.getLogger(__name__)


def load_config(
    index: Index, app_config_file, make_config, make_tasks, *args, **kwargs
):
    app_config_path = Path(app_config_file)
    _, config = next(read_documents(app_config_path))
    config["app_config_file"] = app_config_path.name
    config["task_timestamp"] = int(time.time())

    config = make_config(index, config, **kwargs)

    tasks = make_tasks(index, config, **kwargs)

    return config, iter(tasks)


def pickle_stream(objs, filename) -> int:
    idx = 0
    with open(filename, "wb") as stream:
        for idx, obj in enumerate(objs, start=1):  # noqa: B007
            pickle.dump(obj, stream, pickle.HIGHEST_PROTOCOL)
    return idx


def unpickle_stream(filename):
    with open(filename, "rb") as stream:
        while True:
            try:
                yield pickle.load(stream)
            except EOFError:
                break


def save_tasks(config: dict, tasks, taskfile: str) -> int:
    """Saves the config

    :param config: dict of configuration options common to all tasks
    :param tasks:
    :param taskfile: Name of output file
    :return: Number of tasks saved to the file
    """
    i = pickle_stream(itertools.chain([config], tasks), taskfile)
    if i <= 1:
        # Only saved the config, no tasks!
        os.remove(taskfile)
        return 0
    else:
        _LOG.info("Saved config and %d tasks to %s", i - 1, taskfile)
    return i - 1


def load_tasks(taskfile):
    stream = unpickle_stream(taskfile)
    config = next(stream)
    return config, stream


# This is a function, so it's valid to be lowercase.
#: pylint: disable=invalid-name
app_config_option = click.option(
    "--app-config",
    help="App configuration file",
    type=click.Path(exists=True, readable=True, writable=False, dir_okay=False),
)
#: pylint: disable=invalid-name
load_tasks_option = click.option(
    "--load-tasks",
    "input_tasks_file",
    help="Load tasks from the specified file",
    type=click.Path(exists=True, readable=True, writable=False, dir_okay=False),
)
#: pylint: disable=invalid-name
save_tasks_option = click.option(
    "--save-tasks",
    "output_tasks_file",
    help="Save tasks to the specified file",
    type=click.Path(exists=False),
)
#: pylint: disable=invalid-name
queue_size_option = click.option(
    "--queue-size",
    help="Number of tasks to queue at the start",
    type=click.IntRange(1, 100000),
    default=3200,
)

#: pylint: disable=invalid-name
task_app_options = dc_ui.compose(
    app_config_option,
    load_tasks_option,
    save_tasks_option,
    dc_ui.config_option,
    dc_ui.verbose_option,
    dc_ui.log_queries_option,
)


def _cell_list_from_file(filename):
    cell_matcher = re.compile(r"(-?\d+)(?:\s*(?:,|_|\s)\s*)(-?\d+)")
    with open(filename) as cell_file:
        for line in cell_file:
            match = cell_matcher.match(line)
            if match:
                yield tuple(int(i) for i in match.groups())


def cell_list_to_file(filename, cell_list) -> None:
    with open(filename, "w") as cell_file:
        cell_file.writelines("{},{}\n".format(*cell) for cell in cell_list)


def validate_cell_list(ctx, param, value):
    try:
        if value is None:
            return None
        return list(_cell_list_from_file(value))
    except ValueError:
        raise click.BadParameter(
            'cell_index_list must be a file with lines in the form "14,-11"'
        ) from None


def validate_cell_index(ctx, param, value) -> tuple[int, ...] | None:
    try:
        if value is None:
            return None
        return tuple(int(i) for i in value.split(",", 2))
    except ValueError:
        raise click.BadParameter(
            'cell_index must be specified in the form "14,-11"'
        ) from None


def validate_year(ctx, param, value):
    try:
        if value is None:
            return None
        years = [pd.Period(y) for y in value.split("-", 2)]
        return years[0].start_time.to_pydatetime(warn=False), years[
            -1
        ].end_time.to_pydatetime(warn=False)
    except ValueError:
        raise click.BadParameter(
            "year must be specified as a single year (eg 1996) "
            "or as an inclusive range (eg 1996-2001)"
        ) from None


def break_query_into_years(time_query, **kwargs):
    if time_query is None:
        return [kwargs]
    return [
        dict(time=time_range, **kwargs) for time_range in year_splitter(*time_query)
    ]


def year_splitter(start: str, end: str) -> Generator[tuple[str, str]]:
    """
    Produces a list of time ranges based that represent each year in the range.

    `year_splitter('1992', '1993')` returns:

    ::
        [('1992-01-01 00:00:00', '1992-12-31 23:59:59.9999999'),
         ('1993-01-01 00:00:00', '1993-12-31 23:59:59.9999999')]

    :param start: start year
    :param end: end year
    :return Generator[tuple(str, str)]: strings representing the ranges
    """
    start_ts = pd.Timestamp(start)
    end_ts = pd.Timestamp(end)
    for p in pd.period_range(start=start_ts, end=end_ts, freq="A"):
        yield str(p.start_time), str(p.end_time)


#: pylint: disable=invalid-name
cell_index_option = click.option(
    "--cell-index",
    "cell_index",
    help="Limit the process to a particular cell (e.g. 14,-11)",
    callback=validate_cell_index,
    default=None,
)
#: pylint: disable=invalid-name
cell_index_list_option = click.option(
    "--cell-index-list",
    "cell_index_list",
    help="Limit the process to a file of cells indexes (e.g. 14,-11)",
    callback=validate_cell_list,
    default=None,
)
#: pylint: disable=invalid-name
year_option = click.option(
    "--year",
    "time",
    help="Limit the process to a particular year",
    callback=validate_year,
)


def task_app(make_config, make_tasks):
    """
    Create a `Task App` from a function

    Decorates a function
    :param make_config: callable(index, config, **query)
    :param make_tasks: callable(index, config, **kwargs)
    :return:
    """

    def decorate(app_func):
        def with_app_args(
            index,
            app_config=None,
            input_tasks_file=None,
            output_tasks_file=None,
            *args,
            **kwargs,
        ):
            if (app_config is None) == (input_tasks_file is None):
                click.echo("Must specify exactly one of --app-config, --load-tasks")
                click.get_current_context().exit(1)

            if app_config is not None:
                config, tasks = load_config(
                    index, app_config, make_config, make_tasks, *args, **kwargs
                )
            else:
                config, tasks = load_tasks(input_tasks_file)

            if output_tasks_file:
                num_tasks_saved = save_tasks(config, tasks, output_tasks_file)
                return num_tasks_saved != 0

            return app_func(index, config, tasks, *args, **kwargs)

        return functools.update_wrapper(with_app_args, app_func)

    return decorate


def check_existing_files(paths: Sequence[str | Path]) -> None:
    """Check for existing files and optionally delete them.

    :param paths: sequence of path strings or path objects
    """
    click.echo("Files to be created:")
    existing_files = []
    total = 0
    for path in paths:
        total += 1
        file_path = Path(path)
        file_info = ""
        if file_path.exists():
            existing_files.append(file_path)
            file_info = " - ALREADY EXISTS"
        click.echo(f"{path}{file_info}")

    if existing_files and click.confirm(
        f"There were {len(existing_files)} existing files found "
        "that are not indexed. Delete those files now?"
    ):
        for file_path in existing_files:
            file_path.unlink()

    click.echo(
        f"{total} tasks files to be created ({total - len(existing_files)} "
        f"valid files, {len(existing_files)} existing paths)"
    )


def do_nothing(result) -> None:
    pass


def _wrap_impl(f, args, kwargs, task):
    """
    Helper method, needs to be at the top level
    """
    return f(task, *args, **kwargs)


def wrap_task(f, *args, **kwargs):
    """
    Turn function `f(task, *args, **kwargs)` into `g(task)` in pickle-able fashion
    """
    return functools.partial(_wrap_impl, f, args, kwargs)



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2025 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
"""remove DatasetLocation table

Revision ID: d27eed82e1f6
Revises: 610f32dca3cb
Create Date: 2024-11-13 02:39:55.671819

"""
from typing import Sequence, Union

from alembic import op
import sqlalchemy as sa
from sqlalchemy import Column, Integer, String, DateTime, Text, ForeignKey, UniqueConstraint, Index
from sqlalchemy.dialects import postgresql as postgres
from sqlalchemy.sql import func
from datacube.drivers.postgis._core import METADATA
from sqlalchemy.exc import ProgrammingError


# revision identifiers, used by Alembic.
revision: str = 'd27eed82e1f6'
down_revision: Union[str, None] = '610f32dca3cb'
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    try:
        op.add_column("dataset",
                      Column("uri_scheme", String, comment="The scheme of the uri."),
                      schema="odc")
        op.add_column("dataset",
                      Column("uri_body", String, comment="The body of the uri."),
                      schema="odc")
    except ProgrammingError:
        print("Columns uri_scheme and uri_body already exist in dataset table.")
    # select first active location from DatasetLocation and insert into Dataset
    conn = op.get_bind()
    conn.execute(
        sa.text("""UPDATE odc.dataset d
                SET
                    uri_scheme = subquery.uri_scheme,
                    uri_body = subquery.uri_body
                FROM (
                    SELECT DISTINCT ON (l.dataset_ref)
                        l.dataset_ref, l.uri_scheme, l.uri_body
                    FROM odc.location l
                    WHERE archived IS NULL
                    ORDER BY l.dataset_ref, l.added
                ) subquery
                WHERE d.id = subquery.dataset_ref;""")
    )

    op.drop_table("location", schema="odc", if_exists=True)


def downgrade() -> None:
    loc = op.create_table(
        "location",
        Column("id", Integer, primary_key=True, autoincrement=True),
        Column("dataset_ref", postgres.UUID(as_uuid=True), ForeignKey("odc.dataset.id"), nullable=False,
               comment="The product this dataset belongs to"),
        Column("uri_scheme", String, nullable=False, comment="The scheme of the uri."),
        Column("uri_body", String, nullable=False, comment="The body of the uri."),
        Column("added", DateTime(timezone=True), server_default=func.now(), nullable=False, comment="when added"),
        Column("added_by", Text, server_default=func.current_user(), nullable=False, comment="added by whom"),
        Column("archived", DateTime(timezone=True), default=None, nullable=True, index=True,
               comment="when archived, null for the active location"),
        UniqueConstraint('uri_scheme', 'uri_body', 'dataset_ref'),
        Index("ix_loc_ds_added", "dataset_ref", "added"),
        METADATA,
        schema="odc",
        comment="Where data for the dataset can be found (uri).",
        if_not_exists=True,
    )

    conn = op.get_bind()
    res = conn.execute(
        sa.text("SELECT id, uri_scheme, uri_body FROM odc.dataset WHERE uri_body IS NOT NULL")
    ).fetchall()
    values = [{"dataset_ref": r.id, "uri_scheme": r.uri_scheme, "uri_body": r.uri_body} for r in res]
    op.bulk_insert(loc, values)

    op.drop_column("dataset", "uri_scheme", schema="odc")
    op.drop_column("dataset", "uri_body", schema="odc")



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2025 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
"""
Module
"""

from datacube.index import Index
from datacube.ui.task_app import task_app, wrap_task


def make_test_config(index: Index, config, **kwargs):
    assert index == "Fake Index"
    assert "config_arg" in kwargs

    config["some_item"] = "make_test_config"
    config["num_tasks"] = config.get("num_tasks", 3)
    return config


def make_test_tasks(index: Index, config, **kwargs):
    assert index == "Fake Index"
    assert "task_arg" in kwargs

    num_tasks = config["num_tasks"]
    for i in range(0, num_tasks):
        yield f"Task: {i}"


@task_app(make_config=make_test_config, make_tasks=make_test_tasks)
def my_test_app(index: Index, config, tasks, **kwargs) -> None:
    assert index == "Fake Index"
    assert config["some_item"] == "make_test_config"
    assert "app_arg" in kwargs

    task_list = list(tasks)
    assert len(task_list) == config["num_tasks"]


def test_task_app(tmpdir) -> None:
    index = "Fake Index"

    app_config = tmpdir.join("app_config.yaml")
    app_config.write(
        "name: Test Config\r\ndescription: This is my test app config file"
    )

    my_test_app(index, str(app_config), app_arg=True, config_arg=True, task_arg=True)


def test_task_app_with_task_file(tmpdir) -> None:
    index = "Fake Index"

    app_config = tmpdir.join("app_config.yaml")
    app_config.write(
        "name: Test Config\r\ndescription: This is my test app config file"
    )

    taskfile = tmpdir.join("tasks.bin")
    assert not taskfile.check()

    my_test_app(
        index,
        app_config=str(app_config),
        output_tasks_file=str(taskfile),
        config_arg=True,
        task_arg=True,
    )

    assert taskfile.check()

    my_test_app(index, input_tasks_file=str(taskfile), app_arg=True)


def test_task_app_with_no_tasks(tmpdir) -> None:
    index = "Fake Index"

    app_config = tmpdir.join("app_config.yaml")
    app_config.write(
        "name: Test Config\r\n"
        "description: This is my test app config file\r\n"
        "num_tasks: 0"
    )

    taskfile = tmpdir.join("tasks.bin")
    assert not taskfile.check()

    my_test_app(
        index,
        app_config=str(app_config),
        output_tasks_file=str(taskfile),
        config_arg=True,
        task_arg=True,
    )

    assert not taskfile.check()


def test_task_app_year_splitting() -> None:
    import pandas as pd

    from datacube.ui.task_app import break_query_into_years, validate_year

    one_millisecond = pd.Timedelta("1 ms")

    def is_close(ts1, ts2, max_delta=one_millisecond):
        return abs(pd.Timestamp(ts1) - pd.Timestamp(ts2)) < max_delta

    assert validate_year(None, None, None) is None
    year_range = validate_year(None, None, "1996-2004")
    assert is_close(year_range[0], "1996-01-01 00:00:00")
    assert is_close(year_range[1], "2004-12-31 23:59:59.999999999")

    year_range = validate_year(None, None, "2003")
    assert is_close(year_range[0], "2003-01-01 00:00:00")
    assert is_close(year_range[1], "2003-12-31 23:59:59.999999999")

    # Test that a no year range makes a single query
    year_range = None
    query = break_query_into_years(year_range)
    assert len(query) == 1
    assert query[0] == {}

    # Test that a no year range makes a single query with additional params
    year_range = None
    test_cell_index = (11, 12)
    query = break_query_into_years(year_range, cell_index=test_cell_index)
    assert len(query) == 1
    assert query[0] == {"cell_index": test_cell_index}

    # Test that a single year makes a single query
    year_range = ("1996", "1996")
    query = break_query_into_years(year_range)
    assert len(query) == 1
    assert is_close(query[0]["time"][0], "1996-01-01 00:00:00")
    assert is_close(query[0]["time"][1], "1996-12-31 23:59:59.999999999")

    # Test that a multiple years makes multiple queries
    year_range = ("1996", "1997")
    query = break_query_into_years(year_range)
    assert len(query) == 2
    assert is_close(query[0]["time"][0], "1996-01-01 00:00:00")
    assert is_close(query[0]["time"][1], "1996-12-31 23:59:59.999999999")
    assert is_close(query[1]["time"][0], "1997-01-01 00:00:00")
    assert is_close(query[1]["time"][1], "1997-12-31 23:59:59.999999999")

    # Check that additional kwargs can be used in the query
    year_range = ("1996", "1997")
    test_cell_index = (11, 12)
    query = break_query_into_years(year_range, cell_index=test_cell_index)
    assert len(query) == 2
    assert is_close(query[0]["time"][0], "1996-01-01 00:00:00")
    assert is_close(query[0]["time"][1], "1996-12-31 23:59:59.999999999")
    assert query[0]["cell_index"] == test_cell_index
    assert is_close(query[1]["time"][0], "1997-01-01 00:00:00")
    assert is_close(query[1]["time"][1], "1997-12-31 23:59:59.999999999")
    assert query[1]["cell_index"] == test_cell_index


def test_task_app_cell_index(tmpdir) -> None:
    from datacube.ui.task_app import (
        cell_list_to_file,
        validate_cell_index,
        validate_cell_list,
    )

    assert validate_cell_index(None, None, None) is None
    assert validate_cell_index(None, None, "17,-12") == (17, -12)

    cell_list = [(17, 12), (16, -10), (-23, 0)]

    cell_list_file = tmpdir.join("cell_list.txt")
    if cell_list_file.exists():
        cell_list_file.unlink()
    assert not cell_list_file.check()

    cell_list_to_file(str(cell_list_file), cell_list)

    assert cell_list_file.check()

    assert validate_cell_list(None, None, None) is None
    assert validate_cell_list(None, None, str(cell_list_file)) == cell_list


def test_wrap_task() -> None:
    def task_with_args(task, a, b):
        return task, a, b

    assert task_with_args(1, 2, "a") == (1, 2, "a")
    assert wrap_task(task_with_args, "a", "b")(0) == (0, "a", "b")



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2025 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
"""Defines abstract types for IO drivers."""

from abc import ABCMeta, abstractmethod
from collections.abc import Iterable
from concurrent.futures import Future
from typing import Any, TypeAlias

import numpy as np
from affine import Affine
from odc.geo.crs import CRS

from datacube.storage import BandInfo

# pylint: disable=invalid-name,unsubscriptable-object,pointless-statement

FutureGeoRasterReader: TypeAlias = Future["GeoRasterReader"]
FutureNdarray: TypeAlias = Future[np.ndarray]

RasterShape: TypeAlias = tuple[int, int]
RasterWindow: TypeAlias = tuple[slice, slice]


class GeoRasterReader(metaclass=ABCMeta):
    """Abstract base class for dataset reader."""

    @property
    @abstractmethod
    def crs(self) -> CRS | None: ...  # pragma: no cover

    @property
    @abstractmethod
    def transform(self) -> Affine | None: ...  # pragma: no cover

    @property
    @abstractmethod
    def dtype(self) -> np.dtype: ...  # pragma: no cover

    @property
    @abstractmethod
    def shape(self) -> RasterShape: ...  # pragma: no cover

    @property
    @abstractmethod
    def nodata(self) -> int | float | None: ...  # pragma: no cover

    @abstractmethod
    def read(
        self, window: RasterWindow | None = None, out_shape: RasterShape | None = None
    ) -> FutureNdarray: ...  # pragma: no cover


class ReaderDriver(metaclass=ABCMeta):
    """Interface for Reader Driver"""

    @abstractmethod
    def new_load_context(self, bands: Iterable[BandInfo], old_ctx: Any | None) -> Any:
        """Recycle old context if available/possible and create new context.
        ``old_ctx`` won't be used after this call.

        Same context object is passed to all calls to ``open`` function that
        happen within the same ``dc.load``.

        If your driver doesn't need it just return ``None``
        """
        ...  # pragma: no cover

    @abstractmethod
    def open(
        self, band: BandInfo, ctx: Any
    ) -> FutureGeoRasterReader: ...  # pragma: no cover


class ReaderDriverEntry(metaclass=ABCMeta):
    @property
    @abstractmethod
    def protocols(self) -> list[str]: ...  # pragma: no cover

    @property
    @abstractmethod
    def formats(self) -> list[str]: ...  # pragma: no cover

    @abstractmethod
    def supports(self, protocol: str, fmt: str) -> bool: ...  # pragma: no cover

    @abstractmethod
    def new_instance(self, cfg: dict) -> ReaderDriver: ...  # pragma: no cover
